/**
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Konstantin Svirsky aka crome <mailto:c@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.ge.robot;

import by.informatics.jcx.fingy.ge.AnimationSpeed;
import by.informatics.jcx.fingy.ge.GameEnvironment;
import by.informatics.jcx.fingy.lang.Interpreter;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.util.ArrayList;
import java.awt.Component;
import java.awt.event.ComponentAdapter;
import javax.swing.JPanel;

import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 * @author crome
 */
public class RobotGameEnvironment extends GameEnvironment
{
    private final int RIGHTANGLE     = 90;
    private final int ROTATIONDELTA  = 5;
    private final int ANIMATIONDELTA = 10;

    private int defaultRobotX = 1;
    private int defaultRobotY = 1;

    private boolean redrawOnScrolling = true;
    private boolean isDragging        = false;
    private int startX = 0;
    private int startY = 0;
    protected RobotPanel panel;
    protected Robot robot;
    protected RobotRenderer robotRenderer;
    private int winWidth;
    private int winHeight;
    private int cellSize = 55;
    private ArrayList<Flag> flags = new ArrayList<Flag>();
    private ArrayList<Treasure> treasures = new ArrayList<Treasure>();
    private ArrayList<RobotText>  robotTexts  = new ArrayList<RobotText>();

    private int[][] walls;
    private int defaultRobotDirection = Direction.NORTH;

    public RobotGameEnvironment(int winWidth, int winHeight,
        int fieldWidth, int fieldHeight)
    {
        this.winWidth    = winWidth;
        this.winHeight   = winHeight;

        walls = new int[fieldWidth][fieldHeight];
        encloseFieldByEdge(fieldWidth, fieldHeight);

        robot = new Robot(0, 0, 0, Color.GRAY);
        setRobotX(defaultRobotX);
        setRobotY(defaultRobotY);
        setRobotDirection(defaultRobotDirection);

        initPanel();

        robotRenderer = new RobotRenderer(panel, robot, cellSize,
            fieldWidth, fieldHeight, walls);
        robotRenderer.setRobotFlags(flags);
        robotRenderer.setTreasures(treasures);
        robotRenderer.setTextSource(robotTexts);

        setSize(winWidth, winHeight);
    }

    private void encloseFieldByEdge(int w, int h)
    {
        for (int i = 0; i < h; i++)
        {
            walls[0][i] = walls[0][i] | Direction.WEST;
            walls[w - 1][i] = walls[w - 1][i] | Direction.EAST;
        }
        for (int i = 0; i < w; i++)
        {
            walls[i][0] = walls[i][0] | Direction.NORTH;
            walls[i][h - 1] = walls[i][h - 1] | Direction.SOUTH;
        }
    }

    private void initPanel()
    {
        panel = new RobotPanel(winWidth, winHeight);
        panel.setVisible(true);
        panel.addMouseListener(new MouseAdapter()
        {
            @Override
            public void mouseReleased(MouseEvent e)
            {
                if (e.getButton() == MouseEvent.BUTTON3)
                {
                    isDragging = false;
                    if (redrawOnScrolling)
                    {
                        return;
                    }
                    scrollViewableRegionBy(startX - e.getX(), startY - e.getY());
                    robotRenderer.drawScene();
                    panel.repaint();
                }
            }

            @Override
            public void mousePressed(MouseEvent e)
            {
                if (e.getButton() == MouseEvent.BUTTON3)
                {
                    isDragging = true;
                    startX = e.getX();
                    startY = e.getY();
                }
            }
        });

        panel.addMouseMotionListener(new MouseMotionAdapter()
        {
            @Override
            public void mouseMoved(MouseEvent e)
            {
                robotRenderer.setAxisXY(e.getX(), e.getY());
                robotRenderer.drawScene();
                panel.repaint();
            }

            @Override
            public void mouseDragged(MouseEvent e)
            {
                robotRenderer.setAxisXY(e.getX(), e.getY());
                if (redrawOnScrolling && isDragging)
                {
                    scrollViewableRegionBy(startX - e.getX(), startY - e.getY());
                    startX = e.getX();
                    startY = e.getY();
                }
                robotRenderer.drawScene();
                panel.repaint();
            }
        });

        panel.addComponentListener(new ComponentAdapter()
        {
            @Override
            public void componentResized(ComponentEvent e)
            {
                int pw = panel.getWidth();
                int ph = panel.getHeight();

                setSizeBy(pw - winWidth, ph - winHeight);
                //centeringRobot();
                robotRenderer.drawScene();
            }
        });
    }

    public void updateScene()
    {
        robotRenderer.renderScene();
        panel.repaint();
    }

    public void resetWalls()
    {
        int h = robotRenderer.getFieldHeight();
        int w = robotRenderer.getFieldWidth();

        for (int i = 0; i < h; i++)
        {
            for (int j = 0; j < w; j++)
            {
                walls[j][i] = 0;
            }
        }
    }

    public void addWall(int x, int y, int wall)
    {
        if (x > 1 && (wall & Direction.WEST) == Direction.WEST)
        {
            walls[x - 2][y - 1] = walls[x - 2][y - 1] | Direction.EAST;
            walls[x - 1][y - 1] = walls[x - 1][y - 1] | Direction.WEST;
        }
        if (y > 1 && (wall & Direction.NORTH) == Direction.NORTH)
        {
            walls[x - 1][y - 2] = walls[x - 1][y - 2] | Direction.SOUTH;
            walls[x - 1][y - 1] = walls[x - 1][y - 1] | Direction.NORTH;
        }
        if (x < robotRenderer.getFieldWidth() && (wall & Direction.EAST) == Direction.EAST)
        {
            walls[x][y - 1] = walls[x][y - 1] | Direction.WEST;
            walls[x - 1][y - 1] = walls[x - 1][y - 1] | Direction.EAST;
        }
        if (y < robotRenderer.getFieldHeight() && (wall & Direction.SOUTH) == Direction.SOUTH)
        {
            walls[x - 1][y] = walls[x - 1][y] | Direction.NORTH;
            walls[x - 1][y - 1] = walls[x - 1][y - 1] | Direction.SOUTH;
        }

    }

    public void centeringRobot()
    {
        setViewableRegion((int)robot.getX() - panel.getWidth() / 2,
            (int)robot.getY() - panel.getHeight() / 2);
    }

    public void setViewableRegion(int x, int y)
    {
        robotRenderer.setXView(x);
        robotRenderer.setYView(y);
    }

    public void scrollViewableRegionBy(int x, int y)
    {
        robotRenderer.setXView(robotRenderer.getXView() + x);
        robotRenderer.setYView(robotRenderer.getYView() + y);
    }

    public JPanel getPanel()
    {
        return panel;
    }

    public int getRobotX()
    {
        return (int) robot.getX() / cellSize;
    }

    public int getRobotY()
    {
        return (int) robot.getY() / cellSize;
    }

    public int getWallsAt(int x, int y)
    {
        return walls[x - 1][y - 1];
    }

    public int getRobotDirection()
    {
        return Direction.getDirectionByAngle((360 - (int) robot.getAngle()) % 360);
    }

    public void goForward()
    {
        if (isWallInFront())
        {
            return;
        }
        goForwardWithAnimation(cellSize);

        int x = getRobotX();
        int y = getRobotY();

        for (int i = 0; i < treasures.size(); i++)
        {
            Treasure t = treasures.get(i);
            if (t.getX() == x && t.getY() == y && !t.isFound())
            {
                t.setFound(true);
                updateScene();
                return;
            }
        }
    }

    public boolean isWallInFront()
    {
        int x = getRobotX();
        int y = getRobotY();

        return (walls[x - 1][y - 1] & getRobotDirection()) == getRobotDirection();
    }

    public void goForwardWithAnimation(int distance)
    {
        int i = 0;
        int delta = Math.round(ANIMATIONDELTA * Math.signum(distance));
        RobotTrack track = null;

        if (getAnimationSpeed() == AnimationSpeed.DISABLED)
        {
            delta = distance;
        }

        while (Math.abs(i) < Math.abs(distance))
        {
            i += delta;
            if (Math.abs(i) <= Math.abs(distance))
            {
                robot.goForward(delta);
            }
            else
            {
                robot.goForward(distance - i + delta);
            }

            if (track != null)
            {
                track.setX2(robot.getX());
                track.setY2(robot.getY());
            }
            animateSceneState();
        }
    }

    private void animateSceneState()
    {
        updateScene();
        try
        {
            Thread.sleep(getAnimationSpeed().value);
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }
    }

    public void turnClockwise()
    {
        turnByAngle(RIGHTANGLE);
    }

    public void turnCounterclockwise()
    {
        turnByAngle(-RIGHTANGLE);
    }

    private void turnByAngle(double angle)
    {
        double i = 0;
        double delta = ROTATIONDELTA * Math.signum(angle);

        if (getAnimationSpeed() == AnimationSpeed.DISABLED)
        {
            delta = angle;
        }

        while (Math.abs(i) < Math.abs(angle))
        {
            i += delta;
            if (Math.abs(i) <= Math.abs(angle))
            {
                robot.turnClockwise(delta);
            }
            else
            {
                robot.turnClockwise(angle - i + delta);
            }
            animateSceneState();
        }
    }

    public void drawPicture(String fileName, double x, double y)
    {
        robotRenderer.drawImage(fileName, (int) Math.round(x), (int) Math.round(y));
        updateScene();
    }

    public void setBackgroundColor(Color c)
    {
        robotRenderer.setBackgroundColor(c);
        updateScene();
    }

    public void setTrackColor(Color color)
    {
        robot.setTrackColor(color);
    }

    public Color getTrackColor()
    {
        return robot.getTrackColor();
    }

    public void liftUpBelly()
    {
        robot.setBellyToDown(false);
    }

    public void putFlag(int number)
    {
        int x = getRobotX();
        int y = getRobotY();

        for (int i = 0; i < flags.size(); i++)
        {
            Flag f = flags.get(i);
            if (f.getX() == x && f.getY() == y)
            {
                f.setNumber(number);
                updateScene();
                return;
            }
        }

        flags.add(new Flag(x, y, number));
        updateScene();
    }

    public void takeFlag()
    {
        int x = getRobotX();
        int y = getRobotY();

        for (int i = 0; i < flags.size(); i++)
        {
            Flag f = flags.get(i);

            if (f.getX() == x && f.getY() == y)
            {
                flags.remove(i);
                updateScene();
                return;
            }
        }
    }

    public void setSize(int width, int height)
    {
        panel.setPreferredSize(new Dimension(width, height));
        this.winWidth = width;
        this.winHeight = height;
    }

    public void setSizeBy(int deltaWidth, int deltaHeight)
    {
        setSize(winWidth + deltaWidth, winHeight + deltaHeight);
    }

    @Override
    public Component getVisualComponent()
    {
        return panel;
    }

    @Override
    public void initInterpreter(Interpreter interpreter)
    {
        for (RobotFunction func: RobotFunction.values())
        {
            interpreter.addFunction(func.name, new RobotFunctionInitializer(this, func));
        }
        for (RobotConstants cons: RobotConstants.values())
        {
            interpreter.addValue(cons.name, cons.value);
        }
    }

    @Override
    public void resetState()
    {
        flags.clear();
        robotTexts.clear();
        setAllTreasuresNotFound();
        setRobotX(defaultRobotX);
        setRobotY(defaultRobotY);
        setRobotDirection(defaultRobotDirection);
        robot.setTrackColor(Color.GRAY);
        robot.setBellyToDown(true);
        robotRenderer.clearScene();
        //centeringRobot();
        updateScene();
    }

    private void reinitEnvironment(int width, int height)
    {
        int[][] w = new int[width][height];
        for (int i = 0; i < height; i++)
        {
            for (int j = 0; j < width; j++)
            {
                w[j][i] = walls[j][i];
            }
        }
        walls = w;
        encloseFieldByEdge(width, height);
        robotRenderer.setWalls(walls);

        for (int i = 0; i < treasures.size(); i++)
        {
            Treasure t = treasures.get(i);

            if (t.getX() > width || t.getY() > height)
            {
                treasures.remove(i);
            }
        }

        for (int i = 0; i < flags.size(); i++)
        {
            Flag f = flags.get(i);

            if (f.getX() > width || f.getY() > height)
            {
                flags.remove(i);
            }
        }
    }

    public void setFieldWidth(int width)
    {
        robotRenderer.setFieldWidth(width);
        reinitEnvironment(width, robotRenderer.getFieldHeight());
    }

    public int getFieldWidth()
    {
        return robotRenderer.getFieldWidth();
    }

    public void setFieldHeight(int height)
    {
        robotRenderer.setFieldHeight(height);
        reinitEnvironment(robotRenderer.getFieldWidth(), height);
    }

    public int getFieldHeight()
    {
        return robotRenderer.getFieldHeight();
    }

    public void setRobotDirection(int direction)
    {
        robot.setAngle(360 - Direction.getAngleByDirection(direction));
    }

    public void addTextMessage(RobotText t)
    {
        robotTexts.add(t);
    }

    @Override
    public void loadFromXML(Node node)
    {
        if (node.getNodeName().equals("GameEnvironment") && node.getAttributes().
            getNamedItem("name").getNodeValue().equals("Robot")) {
            resetWalls();
            treasures.clear();
            flags.clear();
            NodeList nodeList = node.getChildNodes();
            for (int i = 0; i < nodeList.getLength(); i++)
            {
                Node item = nodeList.item(i);
                if (item.getNodeName().equals("Robot"))
                {
                    NamedNodeMap a = item.getAttributes();
                    int d = Integer.parseInt(a.getNamedItem("direction").getNodeValue());
                    int x = Integer.parseInt(a.getNamedItem("x").getNodeValue());
                    int y = Integer.parseInt(a.getNamedItem("y").getNodeValue());
                    setRobotX(x);
                    setRobotY(y);
                    setRobotDirection(d);
                }
                else if (item.getNodeName().equals("Wall"))
                {
                    NamedNodeMap a = item.getAttributes();
                    int b = Integer.parseInt(a.getNamedItem("border").getNodeValue());
                    int x = Integer.parseInt(a.getNamedItem("x").getNodeValue());
                    int y = Integer.parseInt(a.getNamedItem("y").getNodeValue());
                    addWall(x, y, b);
                }
                else if (item.getNodeName().equals("Treasure"))
                {
                    NamedNodeMap a = item.getAttributes();
                    boolean b = Boolean.parseBoolean(a.getNamedItem("found").getNodeValue());
                    int     x = Integer.parseInt(a.getNamedItem("x").getNodeValue());
                    int     y = Integer.parseInt(a.getNamedItem("y").getNodeValue());
                    addTreasure(x, y, b);
                }
                else if (item.getNodeName().equals("Field"))
                {
                    NamedNodeMap a = item.getAttributes();
                    int w = Integer.parseInt(a.getNamedItem("width").getNodeValue());
                    int h = Integer.parseInt(a.getNamedItem("height").getNodeValue());
                    setFieldHeight(h);
                    setFieldWidth(w);
                }
            }
            updateScene();
        }
    }

    public int getCellSize()
    {
        return cellSize;
    }

    public void setRobotX(int x)
    {
        robot.setX(x * cellSize);
    }

    public void setRobotY(int y)
    {
        robot.setY(y * cellSize);
    }

    @Override
    public Element convertToXML()
    {
        return null;
    }

    public int flagCount()
    {
        return flags.size();
    }

    public Flag getFlag(int i)
    {
        return flags.get(i);
    }

    public int getFlagNumber()
    {
        int x = getRobotX();
        int y = getRobotY();

        for (int i = 0; i < flags.size(); i++)
        {
            Flag f = flags.get(i);

            if (f.getX() == x && f.getY() == y)
            {
                return f.getNumber();
            }
        }
        return 0;
    }

    @Override
    public ArrayList<String> getFunctionList()
    {
        ArrayList<String> l = new ArrayList<String>();

        for (RobotFunction func : RobotFunction.values())
        {
            l.add(func.name);
        }

        return l;
    }

    public void addTreasure(int x, int y, boolean found)
    {
        for (int i = 0; i < treasures.size(); i++)
        {
            Treasure t = treasures.get(i);
            if (t.getX() == x && t.getY() == y)
            {
                t.setFound(found);
                return;
            }
        }
        treasures.add(new Treasure(x, y, found));
    }

    public int getTreasureCount()
    {
        return treasures.size();
    }

    public void resetTreasures()
    {
        treasures.clear();
    }

    public void setAllTreasuresNotFound()
    {
        for (int i = 0; i < treasures.size(); i++)
        {
            treasures.get(i).setFound(false);
        }
    }

    public int getFoundTreasureCount()
    {
        int s = 0;
        for (int i = 0; i < treasures.size(); i++)
        {
            if (treasures.get(i).isFound())
            {
                s++;
            }
        }
        return s;
    }

    @Override
    public ArrayList<String> getConstantList()
    {
        ArrayList<String> l = new ArrayList<String>();

        for (RobotConstants cons : RobotConstants.values())
        {
            l.add(cons.name);
        }

        return l;
    }

    @Override
    public void purgeInterpreter(Interpreter interpreter)
    {
        for (RobotFunction func: RobotFunction.values())
        {
            interpreter.removeFunction(func.name);
        }
        for (RobotConstants cons: RobotConstants.values())
        {
            interpreter.removeValue(cons.name);
        }
    }
}
