/**
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Konstantin Svirsky aka crome <mailto:c@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.ge.robot;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Stroke;
import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.io.File;
import java.io.IOException;
import java.lang.reflect.InvocationTargetException;

import javax.imageio.ImageIO;
import javax.swing.SwingUtilities;

public class RobotRenderer
{
    private String robotImageFileName = "res/robot.png";

    private Color backColor        = Color.WHITE;
    private Color axisColor        = Color.GRAY;
    private Color fieldCellColor   = Color.GRAY;
    private Color wallColor        = new Color(0xCC6600);
    private Color flagColor        = new Color(0xFF6666);
    private Color treasureColor    = new Color(0xFFCC33);
    private Color transparentColor = new Color(0, 0, 0, 0);

    private BasicStroke axisStroke = new BasicStroke(1,
        BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND, 1, new float[] {3}, 0);

    private Font axisFont = new Font(null, Font.PLAIN, 12);
    private Font flagFont = new Font(null, Font.PLAIN, 10);

    private Robot robot;
    private RobotPanel panel;

    private ArrayList<Flag> flags;
    private ArrayList<Treasure> treasures;
    private ArrayList<RobotText> texts;
    private int[][] walls;

    private int xOffset;
    private int yOffset;

    private int xViewOffset = 0;
    private int yViewOffset = 0;

    private int xAxis;
    private int yAxis;

    private Graphics2D backG;
    private Graphics2D imagesG;
    private Graphics2D tracksG;

    private BufferedImage backgroundImage;
    private BufferedImage tracksImage;
    private BufferedImage robotImage;

    private int maxWidth;
    private int maxHeight;

    private int fieldWidth;
    private int fieldHeight;

    private int cellSize;

    public RobotRenderer(RobotPanel p, Robot t, int cellSize,
        int fieldWidth, int fieldHeight, int[][] walls)
    {
        this.fieldWidth  = fieldWidth;
        this.fieldHeight = fieldHeight;
        this.cellSize    = cellSize;
        this.walls       = walls;

        panel     = p;
        backG     = p.getGraphics2D();
        robot     = t;

        maxWidth  = panel.getMaxWidth();
        maxHeight = panel.getMaxHeight();

        xOffset   = cellSize; //maxWidth  / 2;
        yOffset   = cellSize; //maxHeight / 2;

        backgroundImage = new BufferedImage(maxWidth, maxHeight, BufferedImage.TYPE_INT_ARGB);

        imagesG = backgroundImage.createGraphics();
        imagesG.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
            RenderingHints.VALUE_ANTIALIAS_ON);

        tracksImage = new BufferedImage(maxWidth, maxHeight, BufferedImage.TYPE_INT_ARGB);

        tracksG = tracksImage.createGraphics();
        tracksG.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
            RenderingHints.VALUE_ANTIALIAS_ON);

        try
        {
            robotImage = ImageIO.read(new File(robotImageFileName ));
        }
        catch(IOException e)
        {
            System.out.println("Can not load resource '" + robotImageFileName + "'.");
            e.printStackTrace();
        }
    }

    public void drawRobot()
    {
        int x = (int) robot.getX() - xViewOffset;
        int y = (int) robot.getY() - yViewOffset;
        int w = robotImage.getWidth();
        int h = robotImage.getHeight();

        Graphics2D robotG = (Graphics2D)robotImage.getGraphics();
        AffineTransform xForm = robotG.getTransform();

        xForm.rotate(Math.toRadians(robot.getAngle() + 90), w / 2, h / 2);
        BufferedImage b = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
        Graphics2D g = (Graphics2D) b.getGraphics();
        g.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
            RenderingHints.VALUE_ANTIALIAS_ON);
        g.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
            RenderingHints.VALUE_INTERPOLATION_BILINEAR);

        g.drawImage(robotImage, xForm, null);

        int x_pivot = (x - w / 2);
        int y_pivot = (y - h / 2);

        backG.drawImage(b, x_pivot, y_pivot, null);
    }

    public void setTextSource(ArrayList<RobotText> robotTexts)
    {
        texts = robotTexts;
    }

    public void clearScene()
    {
        cleanDrawingArea();

        tracksG.setBackground(transparentColor);
        tracksG.clearRect(0, 0, maxWidth, maxHeight);

        imagesG.setBackground(transparentColor);
        imagesG.clearRect(0, 0, maxWidth, maxHeight);

        xViewOffset = 0;
        yViewOffset = 0;
    }

    public void drawImage(String fileName, int x, int y)
    {
        try
        {
            BufferedImage image = ImageIO.read(new File(fileName));
            imagesG.drawImage(image, null, x + xOffset, y + yOffset);
        }
        catch(IOException e)
        {
            System.out.println("Can not load resource '" + fileName + "'.");
            e.printStackTrace();
        }
    }

    public void setBackgroundColor(Color c)
    {
        backColor = c;
    }

    public void setXView(int x)
    {
        int w = panel.getWidth();

        xViewOffset = x;

        if (xViewOffset < -maxWidth / 2)
        {
            xViewOffset = -maxWidth / 2;
        }
        if (xViewOffset + w > maxWidth / 2)
        {
            xViewOffset = maxWidth / 2 - w;
        }
    }

    public int getXView()
    {
        return xViewOffset;
    }

    public void setYView(int y)
    {
        int h = panel.getHeight();

        yViewOffset = y;

        if (yViewOffset < -maxHeight / 2)
        {
            yViewOffset = -maxHeight / 2;
        }
        if (yViewOffset + h > maxHeight / 2)
        {
            yViewOffset = maxHeight / 2 - h;
        }
    }

    public int getYView()
    {
        return yViewOffset;
    }

    private void cleanDrawingArea()
    {
        backG.setBackground(backColor);
        backG.clearRect(0, 0, maxWidth, maxHeight);
    }

    public void drawAxis()
    {
        backG.setColor(axisColor);
        Stroke s = backG.getStroke();
        backG.setStroke(axisStroke);
        backG.drawLine(0, yAxis, panel.getWidth(), yAxis);
        backG.drawLine(xAxis, 0, xAxis, panel.getHeight());
        backG.setStroke(s);
        Font f = backG.getFont();
        backG.setFont(axisFont);
        backG.drawString(String.format("[ %1d; %2d]",  xAxis + xViewOffset, -yViewOffset - yAxis),
            xAxis + 10, yAxis - 10);
        backG.setFont(f);
    }

    private void drawField()
    {
        int w = fieldWidth * cellSize;
        int h = fieldHeight * cellSize;

        backG.setColor(fieldCellColor);
        for (int i = 0; i < fieldWidth + 1; i++)
        {
            backG.drawLine(- xViewOffset + i * cellSize + cellSize / 2,
                           - yViewOffset + 0 + cellSize / 2,
                           - xViewOffset + i * cellSize + cellSize / 2,
                           - yViewOffset + h + cellSize / 2);
        }
        for (int i = 0; i < fieldHeight + 1; i++)
        {
            backG.drawLine(- xViewOffset + 0 + cellSize / 2,
                           - yViewOffset + i * cellSize + cellSize / 2,
                           - xViewOffset + w + cellSize / 2,
                           - yViewOffset + i * cellSize + cellSize / 2);
        }
   }

    public void drawScene()
    {
        cleanDrawingArea();

        drawField();

        drawFlags();

        drawTreasures();

        drawRobot();

        drawWalls();

        //drawAxis();

        drawText();
    }

    public void setRobotFlags(ArrayList<Flag> flags)
    {
        this.flags = flags;
    }

    private void drawFlags()
    {
        for (int i = 0; i < flags.size(); i++)
        {
            Flag f = flags.get(i);

            drawFlagAt( - xViewOffset + f.getX() * cellSize,
                        - yViewOffset + f.getY() * cellSize,
                        f.getNumber());
        }
    }

    private void drawFlagAt(int x, int y, int number)
    {
        backG.setColor(Color.black);
        backG.drawLine(x - cellSize / 4 - 1,
                       y - cellSize / 3,
                       x - cellSize / 4 - 1,
                       y + cellSize / 4);
        backG.setColor(flagColor);
        backG.fillRect(x - cellSize / 4, y - cellSize / 3, 2 * cellSize / 5, cellSize / 4);
        backG.setColor(Color.black);
        backG.setFont(flagFont);
        backG.drawString(Integer.toString(number), x - cellSize / 6, y + cellSize / 4);
    }

    public void setWalls(int[][] walls)
    {
        this.walls = walls;
    }

    private void drawWalls()
    {
        backG.setColor(wallColor);
        for (int i = 0; i < fieldHeight; i++)
        {
            for (int j = 0; j < fieldWidth; j++)
            {
                if ((walls[j][i] & Direction.NORTH) == Direction.NORTH)
                {
                    backG.drawLine(- xViewOffset + j * cellSize + cellSize / 2,
                        - yViewOffset + i * cellSize + cellSize / 2 + 1,
                        - xViewOffset + (j + 1) * cellSize + cellSize / 2 ,
                        - yViewOffset + i * cellSize + cellSize / 2 + 1);
                }
                if ((walls[j][i] & Direction.SOUTH) == Direction.SOUTH)
                {
                    backG.drawLine(- xViewOffset + j * cellSize + cellSize / 2,
                        - yViewOffset + (i + 1) * cellSize + cellSize / 2 - 1,
                        - xViewOffset + (j + 1) * cellSize + cellSize / 2 ,
                        - yViewOffset + (i + 1) * cellSize + cellSize / 2 - 1);
                }
                if ((walls[j][i] & Direction.WEST) == Direction.WEST)
                {
                    backG.drawLine(- xViewOffset + j * cellSize + cellSize / 2 + 1,
                        - yViewOffset + i * cellSize + cellSize / 2,
                        - xViewOffset + j * cellSize + cellSize / 2 + 1,
                        - yViewOffset + (i + 1) * cellSize + cellSize / 2);
                }
                if ((walls[j][i] & Direction.EAST) == Direction.EAST)
                {
                    backG.drawLine(- xViewOffset + (j + 1) * cellSize + cellSize / 2 - 1,
                        - yViewOffset + i * cellSize + cellSize / 2,
                        - xViewOffset + (j + 1) * cellSize + cellSize / 2 - 1,
                        - yViewOffset + (i + 1) * cellSize + cellSize / 2);
                }
            }
        }
    }

    private void drawText()
    {
        if (texts != null)
        {
            for (int i = 0; i < texts.size(); i++)
            {
                RobotText t = texts.get(i);
                backG.setColor(t.getColor());
                backG.setFont(t.getFont());
                if (t.getCoordinateType() == RobotText.SCREENCOORDINATE)
                {
                    backG.drawString(t.getText(), t.getX(), t.getY());
                }
                else
                {
                    backG.drawString(t.getText(), t.getX() - xViewOffset,
                        t.getY() - yViewOffset);
                }
            }
        }
    }

    public void setAxisXY(int x, int y)
    {
        xAxis = x;
        yAxis = y;
    }

    public void renderScene()
    {
        try
        {
            SwingUtilities.invokeAndWait(new Runnable()
            {
                @Override
                public void run()
                {
                    drawScene();
                }
            });
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }
        catch (InvocationTargetException e)
        {
            e.printStackTrace();
        }
    }

    public void setTreasures(ArrayList<Treasure> treasures)
    {
        this.treasures = treasures;
    }

    private void drawTreasures()
    {
        for (int i = 0; i < treasures.size(); i++)
        {
            Treasure t = treasures.get(i);
            if (!t.isFound())
            {
                drawTreasureAt(- xViewOffset + t.getX() * cellSize,
                               - yViewOffset + t.getY() * cellSize);
            }
        }
    }

    private void drawTreasureAt(int x, int y)
    {
        int d = 2 * cellSize / 5;

        backG.setColor(treasureColor);
        backG.fillOval(x - d / 2, y - d / 2, d, d);
    }

    public void setFieldWidth(int width)
    {
        fieldWidth = width;
    }

    public int getFieldWidth()
    {
        return fieldWidth;
    }

    public void setFieldHeight(int height)
    {
        fieldHeight = height;
    }

    public int getFieldHeight()
    {
        return fieldHeight;
    }
}
