/**
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Konstantin Svirsky aka crome <mailto:c@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.ge.turtle;

import by.informatics.jcx.fingy.ge.AnimationSpeed;
import by.informatics.jcx.fingy.ge.GameEnvironment;
import by.informatics.jcx.fingy.lang.Interpreter;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.awt.Component;
import java.awt.event.ComponentAdapter;
import javax.swing.JPanel;
import javax.swing.event.MouseInputAdapter;
import javax.swing.event.MouseInputListener;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 *
 * @author crome
 */
public class TurtleGameEnvironment extends GameEnvironment
{
    private final int ROTATIONDELTA  = 5;
    private final int ANIMATIONDELTA = 10;
    private boolean redrawOnScrolling = true;
    private boolean isDragging        = false;
    private int startX = 0;
    private int startY = 0;
    protected Turtle turtle;
    protected TurtleRenderer turtleRenderer;
    private int width;
    private int height;
    private int defaultX = 0;
    private int defaultY = 0;
    private int defaultAngle = 0;
    private Color defaultTrackColor = Color.GRAY;

    public TurtleGameEnvironment(int width, int height)
    {
        this.width  = width;
        this.height = height;

        turtle = new Turtle(defaultX, defaultY, defaultAngle, defaultTrackColor);

        turtleRenderer = new TurtleRenderer(turtle, width, height);

        initPanel();

        setSize(width, height);
    }

    private void initPanel()
    {

        MouseInputListener mouseListener = new MouseInputAdapter()
        {
            @Override
            public void mouseReleased(MouseEvent e)
            {
                if (e.getButton() == MouseEvent.BUTTON3)
                {
                    isDragging = false;
                    if (redrawOnScrolling)
                    {
                        return;
                    }
                    scrollViewableRegionBy(startX - e.getX(), startY - e.getY());
                    turtleRenderer.drawScene();
                }
            }

            @Override
            public void mousePressed(MouseEvent e)
            {
                if (e.getButton() == MouseEvent.BUTTON3)
                {
                    isDragging = true;
                    startX = e.getX();
                    startY = e.getY();
                }
            }

            @Override
            public void mouseEntered(MouseEvent e)
            {
                turtleRenderer.setAxisVisible(true);
                turtleRenderer.drawScene();
            }

            @Override
            public void mouseExited(MouseEvent e)
            {
                turtleRenderer.setAxisVisible(false);
                turtleRenderer.drawScene();
            }

            @Override
            public void mouseMoved(MouseEvent e)
            {
                turtleRenderer.setAxisVisible(true);
                turtleRenderer.setAxisXY(e.getX(), e.getY());
                turtleRenderer.drawScene();
            }

            @Override
            public void mouseDragged(MouseEvent e)
            {
                turtleRenderer.setAxisVisible(true);
                turtleRenderer.setAxisXY(e.getX(), e.getY());
                if (redrawOnScrolling && isDragging)
                {
                    scrollViewableRegionBy(startX - e.getX(), - startY + e.getY());
                    startX = e.getX();
                    startY = e.getY();
                }
                turtleRenderer.drawScene();
            }
        };

        turtleRenderer.addMouseListener(mouseListener);
        turtleRenderer.addMouseMotionListener(mouseListener);

        turtleRenderer.addComponentListener(new ComponentAdapter()
        {
            @Override
            public void componentResized(ComponentEvent e)
            {
                int pw = turtleRenderer.getWidth();
                int ph = turtleRenderer.getHeight();

                setSizeBy(pw - width, ph - height);
                centeringTurtle();
                turtleRenderer.drawScene();
            }
        });
    }

    public void centeringTurtle()
    {
        setViewableRegion((int) turtle.getX() - turtleRenderer.getWidth() / 2,
            - (int) turtle.getY() + turtleRenderer.getHeight() / 2);
        turtleRenderer.drawScene();
    }

    public void setViewableRegion(int x, int y)
    {
        turtleRenderer.setXYView(x, y);
    }

    public void scrollViewableRegionBy(int x, int y)
    {
        turtleRenderer.setXYView(turtleRenderer.getXView() + x, turtleRenderer.getYView() + y);
    }

    public JPanel getPanel()
    {
        return turtleRenderer.getPanel();
    }

    public double getTurtleX()
    {
        return turtle.getX();
    }

    public double getTurtleY()
    {
        return - turtle.getY();
    }

    public double getTurtleAngle()
    {
        return (360 - turtle.getAngle()) % 360;
    }

    public void goForward(double distance)
    {
        goForwardWithAnimation(distance);
    }

    public void goForwardWithAnimation(double distance)
    {
        double x1 = turtle.getX();
        double y1 = turtle.getY();

        double i = 0;
        double delta = ANIMATIONDELTA * Math.signum(distance);
        TurtleTrack track = null;

        if (turtle.isBellyDown())
        {
            track = new TurtleTrack(x1, y1, x1, y1, turtle.getTrackColor());
            turtleRenderer.addTurtleTrack(track);
        }

        if (getAnimationSpeed() == AnimationSpeed.DISABLED)
        {
            delta = distance;
        }

        while (Math.abs(i) < Math.abs(distance))
        {
            i += delta;
            if (Math.abs(i) <= Math.abs(distance))
            {
                turtle.goForward(delta);
            }
            else
            {
                turtle.goForward(distance - i + delta);
            }

            if (track != null)
            {
                track.setX2(turtle.getX());
                track.setY2(turtle.getY());
            }
            animateSceneState();
        }
    }

    private void animateSceneState()
    {
        turtleRenderer.renderScene();

        try
        {
            Thread.sleep(getAnimationSpeed().value);
        }
        catch (InterruptedException e)
        {
            e.printStackTrace();
        }
    }

    public void turnClockwise(double angle)
    {
        double i = 0;
        double delta = ROTATIONDELTA * Math.signum(angle);

        if (getAnimationSpeed() == AnimationSpeed.DISABLED)
        {
            delta = angle;
        }

        while (Math.abs(i) < Math.abs(angle))
        {
            i += delta;
            if (Math.abs(i) <= Math.abs(angle))
            {
                turtle.turnClockwise(delta);
            }
            else
            {
                turtle.turnClockwise(angle - i + delta);
            }
            animateSceneState();
        }
    }

    public void turnCounterclockwise(double angle)
    {
        turnClockwise(-angle);
    }

    public void drawPlot(double r)
    {
        turtleRenderer.drawPlot(r);
        turtleRenderer.renderScene();
    }

    public void drawPicture(String fileName, double x, double y)
    {
        turtleRenderer.drawPicture(fileName, (int) Math.round(x), (int) Math.round(y));
        turtleRenderer.renderScene();
    }

    public void setBackgroundColor(Color c)
    {
        turtleRenderer.setBackgroundColor(c);
        turtleRenderer.renderScene();
    }

    public void setTrackColor(Color color)
    {
        turtle.setTrackColor(color);
    }

    public Color getTrackColor()
    {
        return turtle.getTrackColor();
    }

    public void liftUpBelly()
    {
        turtle.setBellyToDown(false);
    }

    public void liftDownBelly()
    {
        turtle.setBellyToDown(true);
    }

    public boolean isBellyDown()
    {
        return turtle.isBellyDown();
    }

    public void setSize(int width, int height)
    {
        turtleRenderer.setPreferredSize(new Dimension(width, height));
        this.width = width;
        this.height = height;
    }

    public void setSizeBy(int deltaWidth, int deltaHeight)
    {
        setSize(width + deltaWidth, height + deltaHeight);
    }

    @Override
    public Component getVisualComponent()
    {
        return getPanel();
    }

    @Override
    public void initInterpreter(Interpreter interpreter)
    {
        for (TurtleFunction func: TurtleFunction.values())
        {
            interpreter.addFunction(func.name, new TurtleFunctionInitializer(this, func));
        }

        for (TurtleConstants cons: TurtleConstants.values())
        {
            interpreter.addValue(cons.name, cons.value);
        }
    }

    @Override
    public void resetState()
    {
        turtleRenderer.clearFigures();
        turtle.setX(defaultX);
        turtle.setY(defaultY);
        turtle.setAngle(defaultAngle);
        turtle.setTrackColor(defaultTrackColor);
        turtle.setBellyToDown(true);
        turtleRenderer.clearScene();
        centeringTurtle();
        turtleRenderer.renderScene();
    }

    public void drawText(String t)
    {
        turtleRenderer.addTurtleText(new TurtleText(t, turtle.getX(), turtle.getY(),
            turtle.getTrackColor(), null));
        turtleRenderer.renderScene();
    }

    public void addTextMessage(TurtleText t)
    {
        turtleRenderer.addTurtleText(t);
        turtleRenderer.renderScene();
    }

    @Override
    public void loadFromXML(Node node)
    {
        if (node.getNodeName().equals("GameEnvironment") && node.getAttributes().
            getNamedItem("name").getNodeValue().equals("Turtle")) {
            NodeList nodeList = node.getChildNodes();
            for (int i = 0; i < nodeList.getLength(); i++) {
                Node item = nodeList.item(i);
                if (item.getNodeName().equals("BackgroundPicture")) {
                    String path = item.getAttributes().getNamedItem("path").getNodeValue();
                    int x = Integer.parseInt(item.getAttributes().getNamedItem("x").getNodeValue());
                    int y = Integer.parseInt(item.getAttributes().getNamedItem("y").getNodeValue());
                    drawPicture(path, x, y);
                }
            }
        }
    }

    @Override
    public Element convertToXML()
    {
        return null;
    }

    @Override
    public ArrayList<String> getFunctionList()
    {
        ArrayList<String> l = new ArrayList<String>();

        for (TurtleFunction func : TurtleFunction.values())
        {
            l.add(func.name);
        }

        return l;
    }

    @Override
    public ArrayList<String> getConstantList()
    {
        ArrayList<String> l = new ArrayList<String>();

        for (TurtleConstants cons : TurtleConstants.values())
        {
            l.add(cons.name);
        }

        return l;
    }

    @Override
    public void purgeInterpreter(Interpreter interpreter)
    {
        for (TurtleFunction func: TurtleFunction.values())
        {
            interpreter.removeFunction(func.name);
        }
        for (TurtleConstants cons: TurtleConstants.values())
        {
            interpreter.removeValue(cons.name);
        }
    }
}
