/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010. Alexey Rusakovich <mailto:exp_sm@yahoo.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.ide;

import java.awt.Color;
import java.awt.Font;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import javax.swing.text.*;

public class FingyDocument extends DefaultStyledDocument {

    private static final long serialVersionUID = 1L;

    public static String FONT_NAME = "Monospaced";
    public static int FONT_SIZE = 12;
    public static Font FONT = new Font(FONT_NAME, Font.PLAIN, FONT_SIZE);

    private static final Color NORMAL_COLOR = Color.BLACK;
    private static final Color COMMENT_COLOR = new Color(0xFFCC00);  // orange
    private static final Color KEYWORD_COLOR = new Color(0x3366CC);  // blue
    private static final Color FUNCTION_COLOR = new Color(0x009966); // green
    private static final Color CONSTANT_COLOR = new Color(0xAA00DD); // purple
    private static final Color QUOTE_COLOR = new Color(0xFF6666);    // red

    private static final Color NORMAL_LINE_COLOR = Color.WHITE;
    private static final Color RUN_LINE_COLOR = Color.GREEN;
    private static final Color BREAK_LINE_COLOR = Color.PINK;

    private DefaultStyledDocument doc;
    private Element rootElement;
    private boolean multiLineComment;
    private MutableAttributeSet normal;
    private MutableAttributeSet keyword;
    private MutableAttributeSet function;
    private MutableAttributeSet constant;
    private MutableAttributeSet comment;
    private MutableAttributeSet quote;
    private MutableAttributeSet normalLine;
    private MutableAttributeSet runLine;
    private MutableAttributeSet breakLine;
    private MutableAttributeSet normalWithBreak;
    private HashSet<String> functions;
    private HashSet<String> constants;
    private HashSet<String> keywords;

    private Set<Integer> breakLines = new HashSet<Integer>();
    private boolean externalChangeAction = true;

    public FingyDocument() {
        doc = this;
        rootElement = doc.getDefaultRootElement();
        putProperty(DefaultEditorKit.EndOfLineStringProperty, "\n");

        normal = new SimpleAttributeSet();
        StyleConstants.setForeground(normal, NORMAL_COLOR);
        StyleConstants.setFontFamily(normal, FONT_NAME);
        StyleConstants.setFontSize(normal, FONT_SIZE);

        setCharacterAttributes(0, 1, normal, true);

        comment = new SimpleAttributeSet();
        StyleConstants.setForeground(comment, COMMENT_COLOR);
        StyleConstants.setItalic(comment, true);

        keyword = new SimpleAttributeSet();
        StyleConstants.setForeground(keyword, KEYWORD_COLOR);

        function = new SimpleAttributeSet();
        StyleConstants.setForeground(function, FUNCTION_COLOR);

        constant = new SimpleAttributeSet();
        StyleConstants.setForeground(constant, CONSTANT_COLOR);

        quote = new SimpleAttributeSet();
        StyleConstants.setForeground(quote, QUOTE_COLOR);

        normalLine = new SimpleAttributeSet();
        StyleConstants.setBackground(normalLine, NORMAL_LINE_COLOR);

        runLine = new SimpleAttributeSet();
        StyleConstants.setBackground(runLine, RUN_LINE_COLOR);

        breakLine = new SimpleAttributeSet();
        StyleConstants.setBackground(breakLine, BREAK_LINE_COLOR);

        normalWithBreak = new SimpleAttributeSet(normal);
        normalWithBreak.addAttributes(breakLine);

        functions = new HashSet<String>();
        constants = new HashSet<String>();
        // TODO: Move out all highlighting logic to allow change highlight on the fly.
        keywords = new HashSet<String>();
        keywords.add("and");
        keywords.add("break");
        keywords.add("by");
        keywords.add("continue");
        keywords.add("pdef");
        keywords.add("def");
        keywords.add("do");
        keywords.add("elif");
        keywords.add("else");
        keywords.add("end");
        keywords.add("false");
        keywords.add("for");
        keywords.add("from");
        keywords.add("if");
        keywords.add("import");
        keywords.add("loop");
        keywords.add("not");
        keywords.add("or");
        keywords.add("break");
        keywords.add("return");
        keywords.add("to");
        keywords.add("true");
        keywords.add("var");
        keywords.add("void");
        keywords.add("while");
        keywords.add("xor");
        keywords.add("then");
        keywords.add("as");
        keywords.add("const");
        keywords.add("times");
        keywords.add("in");
    }

    public boolean isExternalChange()
    {
        return externalChangeAction;
    }

    /*
     *  Override to apply syntax highlighting after the document has been updated
     */
    @Override
    public void insertString(int offset, String str, AttributeSet a)
            throws BadLocationException {
        int linesBefore = rootElement.getElementCount();
        super.insertString(offset, str, a);
        int linesAfter = rootElement.getElementCount();

        try {
            externalChangeAction = false;
            processChangedLines(offset, str.length());
        } finally {
            externalChangeAction = true;
        }

        if (linesBefore != linesAfter) {
            updateLineColors(rootElement.getElementIndex(offset));
        }
    }

    /*
     *  Override to apply syntax highlighting after the document has been updated
     */
    @Override
    public void remove(int offset, int length) throws BadLocationException {
        int linesBefore = rootElement.getElementCount();
        super.remove(offset, length);
        int linesAfter = rootElement.getElementCount();

        try {
            externalChangeAction = false;
            processChangedLines(offset, 0);
        } finally {
            externalChangeAction = true;
        }

        if (linesBefore != linesAfter) {
            updateLineColors(rootElement.getElementIndex(offset));
        }
    }

    /*
     *  Determine how many lines have been changed,
     *  then apply highlighting to each line
     */
    public void processChangedLines(int offset, int length)
            throws BadLocationException {
        String content = doc.getText(0, doc.getLength());

        //  The lines affected by the latest document update

        int startLine = rootElement.getElementIndex(offset);
        int endLine = rootElement.getElementIndex(offset + length);

        //  Make sure all comment lines prior to the start line are commented
        //  and determine if the start line is still in a multi line comment

        //setMultiLineComment(commentLinesBefore(content, startLine));

        //  Do the actual highlighting

        for (int i = startLine; i <= endLine; i++) {
            applyHighlighting(content, i);
        }

        //  Resolve highlighting to the next end multi line delimiter

        if (isMultiLineComment()) {
            commentLinesAfter(content, endLine);
        } else {
            highlightLinesAfter(content, endLine);
        }
    }

    /*
     *  Highlight lines when a multi line comment is still 'open'
     *  (ie. matching end delimiter has not yet been encountered)
     */
    @SuppressWarnings("unused")
    private boolean commentLinesBefore(String content, int line) {
        int offset = rootElement.getElement(line).getStartOffset();

        //  Start of comment not found, nothing to do

        int startDelimiter = lastIndexOf(content, getStartDelimiter(), offset - 2);

        if (startDelimiter < 0) {
            return false;
        }

        //  Matching start/end of comment found, nothing to do

        int endDelimiter = indexOf(content, getEndDelimiter(), startDelimiter);

        if (endDelimiter < offset & endDelimiter != -1) {
            return false;
        }

        //  End of comment not found, highlight the lines

        doc.setCharacterAttributes(startDelimiter, offset - startDelimiter + 1, comment, false);
        return true;
    }

    /*
     *  Highlight comment lines to matching end delimiter
     */
    private void commentLinesAfter(String content, int line) {
        int offset = rootElement.getElement(line).getEndOffset();

        //  End of comment not found, nothing to do

        int endDelimiter = indexOf(content, getEndDelimiter(), offset);

        if (endDelimiter < 0) {
            return;
        }

        //  Matching start/end of comment found, comment the lines

        int startDelimiter = lastIndexOf(content, getStartDelimiter(), endDelimiter);

        if (startDelimiter < 0 || startDelimiter <= offset) {
            doc.setCharacterAttributes(offset, endDelimiter - offset + 1, comment, false);
        }
    }

    /*
     *  Highlight lines to start or end delimiter
     */
    private void highlightLinesAfter(String content, int line)
            throws BadLocationException {
        int offset = rootElement.getElement(line).getEndOffset();

        //  Start/End delimiter not found, nothing to do

        int startDelimiter = indexOf(content, getStartDelimiter(), offset);
        int endDelimiter = indexOf(content, getEndDelimiter(), offset);

        if (startDelimiter < 0) {
            startDelimiter = content.length();
        }

        if (endDelimiter < 0) {
            endDelimiter = content.length();
        }

        int delimiter = Math.min(startDelimiter, endDelimiter);

        if (delimiter < offset) {
            return;
        }

        //	Start/End delimiter found, reapply highlighting

        int endLine = rootElement.getElementIndex(delimiter);

        for (int i = line + 1; i < endLine; i++) {
            Element branch = rootElement.getElement(i);
            Element leaf = doc.getCharacterElement(branch.getStartOffset());
            AttributeSet as = leaf.getAttributes();

            if (as.isEqual(comment)) {
                applyHighlighting(content, i);
            }
        }
    }

    /*
     *  Parse the line to determine the appropriate highlighting
     */
    private void applyHighlighting(String content, int line)
            throws BadLocationException {
        int startOffset = rootElement.getElement(line).getStartOffset();
        int endOffset = rootElement.getElement(line).getEndOffset() - 1;

        int lineLength = endOffset - startOffset;
        int contentLength = content.length();

        if (endOffset >= contentLength) {
            endOffset = contentLength - 1;
        }

        //  check for multi line comments
        //  (always set the comment attribute for the entire line)

//        if (endingMultiLineComment(content, startOffset, endOffset)
//                || isMultiLineComment()
//                || startingMultiLineComment(content, startOffset, endOffset)) {
//            doc.setCharacterAttributes(startOffset, endOffset - startOffset + 1, comment, false);
//            return;
//        }

        //  set normal attributes for the line

        MutableAttributeSet attrs =
                breakLines.contains(line) ? normalWithBreak : normal;
        doc.setCharacterAttributes(startOffset, lineLength, attrs, true);

        //  check for single line comment

        int index = content.indexOf(getSingleLineDelimiter(), startOffset);

        if ((index > -1) && (index <= endOffset)) {
            doc.setCharacterAttributes(index, endOffset - index + 1, comment, false);
            endOffset = index - 1;
        }

        //  check for tokens

        checkForTokens(content, startOffset, endOffset);
    }

    /*
     *  Does this line contain the start delimiter
     */
    @SuppressWarnings("unused")
    private boolean startingMultiLineComment(String content, int startOffset, int endOffset)
            throws BadLocationException {
        int index = indexOf(content, getStartDelimiter(), startOffset);

        if ((index < 0) || (index > endOffset)) {
            return false;
        } else {
            setMultiLineComment(true);
            return true;
        }
    }

    /*
     *  Does this line contain the end delimiter
     */
    @SuppressWarnings("unused")
    private boolean endingMultiLineComment(String content, int startOffset, int endOffset)
            throws BadLocationException {
        int index = indexOf(content, getEndDelimiter(), startOffset);

        if ((index < 0) || (index > endOffset)) {
            return false;
        } else {
            setMultiLineComment(false);
            return true;
        }
    }

    /*
     *  We have found a start delimiter
     *  and are still searching for the end delimiter
     */
    private boolean isMultiLineComment() {
        return multiLineComment;
    }

    private void setMultiLineComment(boolean value) {
        multiLineComment = value;
    }

    /*
     *	Parse the line for tokens to highlight
     */
    private void checkForTokens(String content, int startOffset, int endOffset) {
        while (startOffset <= endOffset) {
            //  skip the delimiters to find the start of a new token

            while (isDelimiter(content.substring(startOffset, startOffset + 1))) {
                if (startOffset < endOffset) {
                    startOffset++;
                } else {
                    return;
                }
            }

            //  Extract and process the entire token

            if (isQuoteDelimiter(content.substring(startOffset, startOffset + 1))) {
                startOffset = getQuoteToken(content, startOffset, endOffset);
            } else {
                startOffset = getOtherToken(content, startOffset, endOffset);
            }
        }
    }

    /*
     *
     */
    private int getQuoteToken(String content, int startOffset, int endOffset) {
        String quoteDelimiter = content.substring(startOffset, startOffset + 1);
        String escapeString = getEscapeString(quoteDelimiter);

        int index;
        int endOfQuote = startOffset;

        //  skip over the escape quotes in this quote

        index = content.indexOf(escapeString, endOfQuote + 1);

        while ((index > -1) && (index < endOffset)) {
            endOfQuote = index + 1;
            index = content.indexOf(escapeString, endOfQuote);
        }

        // now find the matching delimiter

        index = content.indexOf(quoteDelimiter, endOfQuote + 1);

        if ((index < 0) || (index > endOffset)) {
            endOfQuote = endOffset;
        } else {
            endOfQuote = index;
        }

        doc.setCharacterAttributes(startOffset, endOfQuote - startOffset + 1, quote, false);

        return endOfQuote + 1;
    }

    /*
     *
     */
    private int getOtherToken(String content, int startOffset, int endOffset) {
        int endOfToken = startOffset + 1;

        while (endOfToken <= endOffset) {
            if (isDelimiter(content.substring(endOfToken, endOfToken + 1))) {
                break;
            }

            endOfToken++;
        }

        String token = content.substring(startOffset, endOfToken);

        if (isKeyword(token)) {
            doc.setCharacterAttributes(startOffset, endOfToken - startOffset, keyword, false);
        }
        else if (isFunction(token)) {
            doc.setCharacterAttributes(startOffset, endOfToken - startOffset, function, false);
        }
        else if (isConstant(token)) {
            doc.setCharacterAttributes(startOffset, endOfToken - startOffset, constant, false);
        }

        return endOfToken + 1;
    }

    /*
     *  Assume the needle will the found at the start/end of the line
     */
    private int indexOf(String content, String needle, int offset) {
        int index;

        while ((index = content.indexOf(needle, offset)) != -1) {
            String text = getLine(content, index).trim();

            if (text.startsWith(needle) || text.endsWith(needle)) {
                break;
            } else {
                offset = index + 1;
            }
        }

        return index;
    }

    /*
     *  Assume the needle will the found at the start/end of the line
     */
    private int lastIndexOf(String content, String needle, int offset) {
        int index;

        while ((index = content.lastIndexOf(needle, offset)) != -1) {
            String text = getLine(content, index).trim();

            if (text.startsWith(needle) || text.endsWith(needle)) {
                break;
            } else {
                offset = index - 1;
            }
        }

        return index;
    }

    private String getLine(String content, int offset) {
        int line = rootElement.getElementIndex(offset);
        Element lineElement = rootElement.getElement(line);
        int start = lineElement.getStartOffset();
        int end = lineElement.getEndOffset();
        return content.substring(start, end - 1);
    }

    /*
     *  Override for other languages
     */
    protected boolean isDelimiter(String character) {
        String operands = ";:{}()[]+-/%<=>!&|^~*";

        if (Character.isWhitespace(character.charAt(0))
                || operands.indexOf(character) != -1) {
            return true;
        } else {
            return false;
        }
    }

    /*
     *  Override for other languages
     */
    protected boolean isQuoteDelimiter(String character) {
        String quoteDelimiters = "\"'";

        if (quoteDelimiters.indexOf(character) < 0) {
            return false;
        } else {
            return true;
        }
    }

    /*
     *  Override for other languages
     */
    protected boolean isKeyword(String token) {
        return keywords.contains(token);
    }

    protected boolean isFunction(String token) {
        return functions.contains(token);
    }

    protected boolean isConstant(String token) {
        return constants.contains(token);
    }

    public void addFunction(String func) {
        functions.add(func);
    }

    public void addFunctions(Collection<String> func) {
        functions.addAll(func);
    }

    public void removeFunction(String func) {
        functions.remove(func);
    }

    public void removeFunctions(Collection<String> funcs) {
        functions.removeAll(funcs);
    }

    public void addConstant(String con) {
        constants.add(con);
    }

    public void removeConstant(String con) {
        constants.remove(con);
    }

    public void addConstants(Collection<String> cons) {
        constants.addAll(cons);
    }

    public void removeConstants(Collection<String> cons) {
        constants.removeAll(cons);
    }

    /*
     *  Override for other languages
     */
    protected String getStartDelimiter() {
        return "/*";
    }

    /*
     *  Override for other languages
     */
    protected String getEndDelimiter() {
        return "*/";
    }

    /*
     *  Override for other languages
     */
    protected String getSingleLineDelimiter() {
        return "#";
    }

    /*
     *  Override for other languages
     */
    protected String getEscapeString(String quoteDelimiter) {
        return "\\" + quoteDelimiter;
    }

    /*
     *
     */
    protected String addMatchingBrace(int offset) throws BadLocationException {
        StringBuffer whiteSpace = new StringBuffer();
        int line = rootElement.getElementIndex(offset);
        int i = rootElement.getElement(line).getStartOffset();

        while (true) {
            String temp = doc.getText(i, 1);

            if (temp.equals(" ") || temp.equals("\t")) {
                whiteSpace.append(temp);
                i++;
            } else {
                break;
            }
        }

        return "{\n" + whiteSpace.toString() + "\t\n" + whiteSpace.toString() + "}";
    }

    public void updateHighlighting() {
        try {
            try {
                externalChangeAction = false;
                String content = doc.getText(0, doc.getLength());

                int linesCount = rootElement.getElementCount();

                for (int i = 0; i < linesCount; i++) {
                    applyHighlighting(content, i);
                }
            } finally {
                externalChangeAction = true;
            }
        } catch (BadLocationException ex) {
        }
    }

    private void updateLineColors(int startLine) {
        int linesCount = rootElement.getElementCount();

        // update bg's of all lines after
        for (int i = startLine; i < linesCount; i++) {
            setBreakLineState(i, breakLines.contains(i));
        }
    }

    private boolean runLineActive = false;
    private int runLineNum;
    private MutableAttributeSet runLinePrevState;

    void setRunLine(int line) {
        // this check might be unnecessary
        if (runLineActive) {
            // restore previously runned line state
            setLineAttributes(runLineNum, runLinePrevState);
        }

        // remember this line's state
        runLinePrevState = breakLines.contains(line) ? breakLine : normalLine;

        // apply new state
        setLineAttributes(line, runLine);
        runLineNum = line;
        runLineActive = true;
    }

    void disableRunLine() {
        // restore line state
        setLineAttributes(runLineNum, runLinePrevState);
        runLineActive = false;
    }

    void clearBreakLines() {
        breakLines.clear();
        updateLineColors(0);
    }

    void setBreakLine(int line, boolean enabled) {
        if (enabled) {
            breakLines.add(line);
        } else {
            breakLines.remove(line);
        }
        setBreakLineState(line, enabled);
    }

    private void setBreakLineState(int line, boolean enabled) {
        MutableAttributeSet attrs = enabled ? breakLine : normalLine;

        if (runLineActive && line == runLineNum) {
            // if this line is currently being runned, schedule breakLine state
            runLinePrevState = attrs;
        } else {
            // if not, apply immediately
            setLineAttributes(line, attrs);
        }
    }

    private void setLineAttributes(int line, AttributeSet attrs) {
        int startOffset = rootElement.getElement(line).getStartOffset();
        int endOffset = rootElement.getElement(line).getEndOffset();

        try {
            externalChangeAction = false;
            doc.setCharacterAttributes(startOffset, endOffset - startOffset,
                    attrs, false);
        } finally {
            externalChangeAction = true;
        }
    }
}
