/*
 * Fingy. Educational programming environment for children. Copyright (C) 2010.
 * Zhlobich Andrei <mailto:anjensan@informatics.by>
 * 
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.ide;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;

import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JViewport;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.Document;

public class FingySourceEditor extends JPanel {

    private static final long serialVersionUID = 1L;
    private static final int MINIMUM_LINE_DIGITS = 3;
    private static final int HORIZONTAL_PADDING = 4;
    private static final Color NUMBERS_BG_COLOR = new Color(0xEEEEEE);
    private static final Color NUMBERS_FG_COLOR = Color.DARK_GRAY;

    private JFingyPane panel;
    private LinesPanel linesPanel;
    private JPanel sidePane;
    private JComponent leftBottomCorner;
    private JScrollPane scrollPane;
    private JViewport editorViewport;
    private JViewport linesViewport;
    private JScrollBar horizontalScrollBar;

    private int linesPanelCharWidth;
    private int lineHeight;

    private ChangeListener panelsSynchronizer = new ChangeListener() {

        @Override
        public void stateChanged(ChangeEvent e) {
            Point p = editorViewport.getViewPosition();
            linesViewport.setViewPosition(new Point(0, p.y));
        }
    };

    private class LinesPanel extends JPanel {

        private static final long serialVersionUID = 1L;
        private static final int NOLINE = Integer.MIN_VALUE;

        private int yOffset;
        private int drawedLinesCount;

        public LinesPanel() {
            FontMetrics fm = panel.getFontMetrics(panel.getFont());
            yOffset = lineHeight - fm.getDescent();
            addChangeListener(panel.getDocument());
            drawedLinesCount = 0;
        }

        private void initGraphics(Graphics g) {
            Graphics2D g2 = (Graphics2D) g;
            g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING,
                    RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
        }

        private void addChangeListener(Document doc) {
            doc.addDocumentListener(new DocumentListener() {
                @Override
                public void removeUpdate(DocumentEvent e) {
                    update(e);
                }

                @Override
                public void insertUpdate(DocumentEvent e) {
                    update(e);
                }

                @Override
                public void changedUpdate(DocumentEvent e) {
                    // do nothing
                }

                private void update(DocumentEvent e) {
                    invalidateSize();
                }
            });
        }

        private int getLineYCoordinate(int line) {
            if (line < 0) {
                return NOLINE;
            }
            int lpos = panel.getLinePosition(line);
            if (lpos < 0) {
                return NOLINE;
            }
            try {
                return panel.modelToView(lpos).y + yOffset;
            } catch (BadLocationException e1) {
                throw new RuntimeException(e1);
            }
        }

        private void invalidateSize() {
            int lines = panel.getDocument().getDefaultRootElement()
                    .getElementCount();
            if (lines == drawedLinesCount) {
                return;
            }
            int y1 = lineHeight * Math.min(lines, drawedLinesCount);
            int y2 = lineHeight * Math.max(lines, drawedLinesCount);
            repaint(0, y1, getWidth(), y2 - y1);
            drawedLinesCount = lines;
        }

        @Override
        public void paintComponent(Graphics g) {
            super.paintComponent(g);
            initGraphics(g);
            FontMetrics fm = g.getFontMetrics();
            Rectangle clip = g.getClipBounds();
            int start = (int) (clip.getMinY() / lineHeight);
            int end = (int) (clip.getMaxY() / lineHeight + 1);
            int w = getWidth() - HORIZONTAL_PADDING;
            for (int i = start; i <= end; ++i) {
                String s = Integer.toString(i + 1);
                int y = getLineYCoordinate(i);
                if (y != NOLINE) {
                    g.setColor(NUMBERS_BG_COLOR);
                    g.fillRect(0, y - lineHeight, w, lineHeight);
                    g.setColor(NUMBERS_FG_COLOR);
                    g.drawString(s, w - fm.stringWidth(s), y);
                }
            }
        }
    };

    public FingySourceEditor() {
        super();
        initComponents();
    }

    public JFingyPane getTextPanel() {
        return panel;
    }

    public JPanel getLinesPanel() {
        return linesPanel;
    }

    private void initComponents() {
        panel = new JFingyPane();
        FontMetrics fm = panel.getFontMetrics(panel.getFont());
        lineHeight = fm.getHeight();

        linesPanel = new LinesPanel();
        linesPanel.setBackground(NUMBERS_BG_COLOR);
        linesPanel.setForeground(NUMBERS_FG_COLOR);

        scrollPane = new JScrollPane(panel);
        // disable border!
        scrollPane.setBorder(null);
        horizontalScrollBar = scrollPane.getHorizontalScrollBar();

        editorViewport = scrollPane.getViewport();
        editorViewport.addChangeListener(panelsSynchronizer);

        linesViewport = new JViewport();
        linesViewport.setView(linesPanel);

        sidePane = new JPanel();
        leftBottomCorner = new JPanel();
        sidePane.setLayout(new BorderLayout());
        sidePane.add(linesViewport, BorderLayout.CENTER);
        sidePane.add(leftBottomCorner, BorderLayout.SOUTH);

        linesPanel.setFont(panel.getFont());
        fm = linesPanel.getFontMetrics(linesPanel.getFont());
        linesPanelCharWidth = fm.charWidth('8');
        fm = panel.getFontMetrics(panel.getFont());

        // resize `linesPanel` synchronously with `panel`
        panel.addComponentListener(new ComponentAdapter() {

            @Override
            public void componentResized(ComponentEvent e) {
                linesPanel.setPreferredSize(new Dimension(getLinesPanelWidth(),
                        panel.getHeight()));
                leftBottomCorner.setVisible(horizontalScrollBar.isVisible());
                leftBottomCorner.setPreferredSize(new Dimension(0,
                        horizontalScrollBar.getHeight()));
            }
        });

        setLayout(new BorderLayout());
        add(scrollPane, BorderLayout.CENTER);
        add(sidePane, BorderLayout.WEST);
    }

    public void setEditable(boolean editable) {
        panel.setEditable(editable);
    }

    public int getLinePosition(int line) {
        return panel.getLinePosition(line);
    }

    public void setCaretPosition(int pos) {
        panel.setCaretPosition(pos);
    }

    public int getLine() {
        return panel.getLine();
    }

    public int getColumn() {
        return panel.getColumn();
    }

    public FingyDocument getFingyDocument() {
        return panel.getFingyDocument();
    }

    public void setText(String text) {
        panel.setText(text);
    }

    public String getText() {
        return panel.getText();
    }

    public String getText(int offs, int len) throws BadLocationException {
        return panel.getText(offs, len);
    }

    private int getNumberDigitsCount() {
        int x = 0;
        int n = panel.getLinesCount();
        while (n != 0) {
            n /= 10;
            ++x;
        }
        return Math.max(x, MINIMUM_LINE_DIGITS);
    }

    private int getLinesPanelWidth() {
        return getNumberDigitsCount() * linesPanelCharWidth
                + HORIZONTAL_PADDING;
    }
}
