/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010. Alexey Rusakovich <mailto:exp_sm@yahoo.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.ide;

import by.informatics.jcx.fingy.ge.AnimationSpeed;
import by.informatics.jcx.fingy.ge.GameEnvironment;
import by.informatics.jcx.fingy.ge.robot.RobotGameEnvironment;
import by.informatics.jcx.fingy.ge.turtle.TurtleGameEnvironment;
import by.informatics.jcx.fingy.lang.Interpreter;
import by.informatics.jcx.fingy.lang.LangException;
import by.informatics.jcx.fingy.lang.Module;
import by.informatics.jcx.fingy.lang.SyntaxError;
import by.informatics.jcx.fingy.lang.VMContext;
import by.informatics.jcx.fingy.lang.modules.BuiltinModule;
import by.informatics.jcx.fingy.lang.modules.MathModule;
import by.informatics.jcx.fingy.lang.modules.StringModule;
import java.awt.BorderLayout;

import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.util.EventObject;
import java.util.HashSet;
import java.util.Set;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.event.CaretEvent;
import javax.swing.event.CaretListener;
import javax.swing.event.UndoableEditEvent;
import javax.swing.event.UndoableEditListener;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.undo.UndoManager;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import org.jdesktop.application.Action;
import org.jdesktop.application.Application.ExitListener;
import org.jdesktop.application.SingleFrameApplication;
import org.jdesktop.application.FrameView;

/**
 * The application's main frame.
 */
public class FingyView extends FrameView {

    // called in RunThread
    void processRun() {
        setPaused(false);
        setStopped(false);
        setRunning(true);
    }

    // called in RunThread
    void processStop(boolean finished) {
        setRunning(false);
        setStopped(finished);
        setPaused(!finished);
    }

    // called in EDT
    void processRuntimeError(LangException ex) {
        outputPanel.printfError("Runtime error on line %1d: %2s\n",
                ex.getLineNumber(), ex.getMessage());
    }

    // running state properties
    private final static String RUNNING_PROPERTY = "running";
    private final static String NOT_RUNNING_PROPERTY = "notRunning";
    private final static String PAUSED_PROPERTY = "paused";
    private final static String STOPPED_PROPERTY = "stopped";
    private final static String NOT_STOPPED_PROPERTY = "notStopped";
    private final static String GE_WINDOW_LOADED_PROPERTY = "GEWindowLoaded";

    private boolean outPrevVisible = false;
    private boolean gePrevVisible = false;
    private boolean watchPrevVisible = false;
    private FingyDocument fingyDocument;

    protected UndoManager undo = new MyUndoManager();

    public FingyView(SingleFrameApplication app) {
        super(app);

        initComponents();

        fingyPane = new FingySourceEditor();
        fingyPane.setName("jFingyPane");
        fingyPane.getTextPanel().addCaretListener(new CaretListener() {

            @Override
            public void caretUpdate(CaretEvent e) {
                updateStatusLineNumber();
            }
        });

        sourceEditorPanel.setLayout(new BorderLayout());
        sourceEditorPanel.add(fingyPane, BorderLayout.CENTER);

        fingyScriptFileChooser = new JFileChooser(new File("."));
        fingyScriptFileChooser.addChoosableFileFilter(
                new FileNameExtensionFilter("Fingy script", "fingy"));

        gameEnvironmentFileChooser = new JFileChooser(new File("."));
        gameEnvironmentFileChooser.addChoosableFileFilter(
                new FileNameExtensionFilter("Game environment", "xml"));

        appTitle = getContext().getResourceMap(FingyApp.class).getString("Application.title");
        getFrame().setTitle("untitled - " + appTitle);

        fingyDocument = fingyPane.getFingyDocument();
        fingyDocument.addUndoableEditListener(new UndoableEditListener() {
            @Override
            public void undoableEditHappened(UndoableEditEvent e) {
                MyUndoableEdit me;
                if (fingyDocument.isExternalChange()) {
                    me = new MyUndoableEdit(e.getEdit(), true);
                    updateModified(true);
                } else {
                    me = new MyUndoableEdit(e.getEdit(), false);
                }
                undo.addEdit(me);
            }
        });

        updateStatusLineNumber();

        getApplication().addExitListener(new ExitListener() {

            @Override
            public boolean canExit(EventObject event) {
                return canCloseFile();
            }

            @Override
            public void willExit(EventObject event) {
                if (!runner.isFinished()) {
                    runner.stop();
                }
            }
        });

        runner = new RunServer(this);

        outputPanel = new OutputPanel();
        outputWindow = createPanelWindow("Output", outputPanel);

        watchPanel = new WatchPanel(runner);
        watchPanel.updateVariables();
        watchWindow = createPanelWindow("Watches", watchPanel);

        initInterpreter();

        devInit();

        addPropertyChangeListener(STOPPED_PROPERTY, new PropertyChangeListener() {

            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                boolean stopped = (Boolean) evt.getNewValue();
                fingyPane.setEditable(stopped);
            }
        });

        addPropertyChangeListener(PAUSED_PROPERTY, new PropertyChangeListener() {

            @Override
            public void propertyChange(PropertyChangeEvent evt) {
                boolean paused = (Boolean) evt.getNewValue();
                if (paused) {
                    int line = runner.getCurrentLine();
                    debugOutput("paused at line " + line);
                    fingyDocument.setRunLine(line - 1);

                    int pos = fingyPane.getLinePosition(line - 1);
                    fingyPane.setCaretPosition(pos);

                    // TODO: there might be a better way than creating a new list every time
                    watchPanel.updateVariables();
                } else {
                    fingyDocument.disableRunLine();
                    watchPanel.updateVariables();
                }
            }
        });

        // workaround to minimize panel windows with the main one
        // TODO: if ANY normal way exists to achieve it, tell us, for god's sake!
        getFrame().addWindowListener(new WindowAdapter() {

            @Override
            public void windowIconified(WindowEvent e) {
                // on minimizing, remember panels states and hide them
                outPrevVisible = outputWindow.isVisible();
                outputWindow.setVisible(false);

                watchPrevVisible = watchWindow.isVisible();
                watchWindow.setVisible(false);

                if (geWindow != null) {
                    gePrevVisible = geWindow.isVisible();
                    geWindow.setVisible(false);
                }
            }

            @Override
            public void windowDeiconified(WindowEvent e) {
                // on restoring, recover panels states
                outputWindow.setVisible(outPrevVisible);
                watchWindow.setVisible(watchPrevVisible);

                if (geWindow != null) {
                    geWindow.setVisible(gePrevVisible);
                }

                getFrame().toFront();
            }
        });
    }

    private void updateStatusLineNumber() {
        int line = fingyPane.getLine() + 1;
        int col = fingyPane.getColumn() + 1;
        statusLineNumber.setText("Line: " + line + ", Column: " + col);
    }

    private JDialog createPanelWindow(String title, JPanel panel) {
        JDialog dialog = new JDialog();
        dialog.setTitle(title);
        dialog.setContentPane(panel);
        //dialog.setLocationRelativeTo(getFrame()); //useless line if we do arrange
        dialog.pack();

        return dialog;
    }

    private void initInterpreter() {
        isCompiled = false;

        itp = new Interpreter();

        for (IOFunction func : IOFunction.values()) {
            itp.addFunction(func.name, func.getFunction(this));
            fingyDocument.addFunction(func.name);
        }

        Module mathModule = new MathModule();
        itp.addModule("math", mathModule);
        Module strModule = new StringModule();
        itp.addModule("string", strModule);

        // highlight builtin functions
        for (String name : BuiltinModule.getNames()) {
            fingyDocument.addFunction(name);
        }
    }

    protected void arrangeWindows() {
        java.awt.Rectangle r = getFrame().getBounds();
        outputWindow.setBounds(r.x, r.y + r.height,
                r.width, outputWindow.getHeight());
        outputWindow.setVisible(true);
        if (isGEWindowLoaded()) {
            arrangeGEWindow(geWindow);
        }
        watchWindow.setLocation(r.x + r.width, r.y + r.height);
        getFrame().toFront();
    }

    private void arrangeGEWindow(JDialog window) {
        java.awt.Rectangle r = getFrame().getBounds();
        window.setLocation(r.x + r.width, r.y);
        window.setVisible(true);
    }

    /*
     * Performs initializations necessary in the development stage.
     * Should not be run in release builts.
     */
    private void devInit() {
        initGameEnv();
    }

    private void initGameEnv() {
        gameEnv = new TurtleGameEnvironment(500, 500);
        regGE(gameEnv);
        resetGEAction();
        initGEWindow();
        reloadGEWindow("Turtle", true);
        setGEWindowLoaded(true);
        changeAnimationSpeed();
    }

    void debugOutput(String msg) {
        if (spamDebug) {
            boolean inEdt = SwingUtilities.isEventDispatchThread();
            final String out = "[debug] " + (inEdt ? "" : "NOT ") + "in EDT, " + msg;

            System.out.println(out);
        }
    }

    protected OutputPanel getOutputPanel() {
        return outputPanel;
    }

    // TODO: Eclipse warns 'Unnecessary directive'
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    //@SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        mainPanel = new javax.swing.JPanel();
        sourceEditorPanel = new javax.swing.JPanel();
        menuBar = new javax.swing.JMenuBar();
        javax.swing.JMenu fileMenu = new javax.swing.JMenu();
        newMenuItem = new javax.swing.JMenuItem();
        openMenuItem = new javax.swing.JMenuItem();
        loadGEMenuItem = new javax.swing.JMenuItem();
        saveMenuItem = new javax.swing.JMenuItem();
        saveAsMenuItem = new javax.swing.JMenuItem();
        fileSeparator1 = new javax.swing.JPopupMenu.Separator();
        javax.swing.JMenuItem exitMenuItem = new javax.swing.JMenuItem();
        editMenu = new javax.swing.JMenu();
        undoMenuItem = new javax.swing.JMenuItem();
        redoMenuItem = new javax.swing.JMenuItem();
        editSeparator1 = new javax.swing.JPopupMenu.Separator();
        cutMenuItem = new javax.swing.JMenuItem();
        copyMenuItem = new javax.swing.JMenuItem();
        pasteMenuItem = new javax.swing.JMenuItem();
        debugMenu = new javax.swing.JMenu();
        runMenuItem = new javax.swing.JMenuItem();
        stopMenuItem = new javax.swing.JMenuItem();
        pauseMenuItem = new javax.swing.JMenuItem();
        continueMenuItem = new javax.swing.JMenuItem();
        debugSeparator1 = new javax.swing.JPopupMenu.Separator();
        toggleBreakpointMenuItem = new javax.swing.JMenuItem();
        debugSeparator2 = new javax.swing.JPopupMenu.Separator();
        stepOverMenuItem = new javax.swing.JMenuItem();
        stepIntoMenuItem = new javax.swing.JMenuItem();
        settingsMenu = new javax.swing.JMenu();
        animationSpeedMenu = new javax.swing.JMenu();
        asDisabledMenuItem = new javax.swing.JRadioButtonMenuItem();
        asSlowMenuItem = new javax.swing.JRadioButtonMenuItem();
        asNormalMenuItem = new javax.swing.JRadioButtonMenuItem();
        asFastMenuItem = new javax.swing.JRadioButtonMenuItem();
        asVFastMenuItem = new javax.swing.JRadioButtonMenuItem();
        windowMenu = new javax.swing.JMenu();
        showOutputMenuItem = new javax.swing.JMenuItem();
        showGEMenuItem = new javax.swing.JMenuItem();
        showWatchesMenuItem = new javax.swing.JMenuItem();
        windowSeparator1 = new javax.swing.JPopupMenu.Separator();
        clearOutputMenuItem = new javax.swing.JMenuItem();
        resetGEMenuItem = new javax.swing.JMenuItem();
        statusPanel = new javax.swing.JPanel();
        statusMessageLabel = new javax.swing.JLabel();
        statusLineNumber = new javax.swing.JLabel();
        animSpeedGroup = new javax.swing.ButtonGroup();

        mainPanel.setName("mainPanel"); // NOI18N
        mainPanel.setLayout(new java.awt.BorderLayout());

        sourceEditorPanel.setName("sourceEditorPanel"); // NOI18N

        javax.swing.GroupLayout sourceEditorPanelLayout = new javax.swing.GroupLayout(sourceEditorPanel);
        sourceEditorPanel.setLayout(sourceEditorPanelLayout);
        sourceEditorPanelLayout.setHorizontalGroup(
            sourceEditorPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 400, Short.MAX_VALUE)
        );
        sourceEditorPanelLayout.setVerticalGroup(
            sourceEditorPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGap(0, 260, Short.MAX_VALUE)
        );

        mainPanel.add(sourceEditorPanel, java.awt.BorderLayout.CENTER);

        menuBar.setName("menuBar"); // NOI18N

        org.jdesktop.application.ResourceMap resourceMap = org.jdesktop.application.Application.getInstance().getContext().getResourceMap(FingyView.class);
        fileMenu.setText(resourceMap.getString("fileMenu.text")); // NOI18N
        fileMenu.setName("fileMenu"); // NOI18N

        javax.swing.ActionMap actionMap = org.jdesktop.application.Application.getInstance().getContext().getActionMap(FingyView.class, this);
        newMenuItem.setAction(actionMap.get("newAction")); // NOI18N
        newMenuItem.setName("newMenuItem"); // NOI18N
        fileMenu.add(newMenuItem);

        openMenuItem.setAction(actionMap.get("openAction")); // NOI18N
        openMenuItem.setName("openMenuItem"); // NOI18N
        fileMenu.add(openMenuItem);

        loadGEMenuItem.setAction(actionMap.get("loadGEAction")); // NOI18N
        loadGEMenuItem.setName("loadGEMenuItem"); // NOI18N
        fileMenu.add(loadGEMenuItem);

        saveMenuItem.setAction(actionMap.get("saveAction")); // NOI18N
        saveMenuItem.setName("saveMenuItem"); // NOI18N
        fileMenu.add(saveMenuItem);

        saveAsMenuItem.setAction(actionMap.get("saveAsAction")); // NOI18N
        saveAsMenuItem.setName("saveAsMenuItem"); // NOI18N
        fileMenu.add(saveAsMenuItem);

        fileSeparator1.setName("fileSeparator1"); // NOI18N
        fileMenu.add(fileSeparator1);

        exitMenuItem.setAction(actionMap.get("quit")); // NOI18N
        exitMenuItem.setName("exitMenuItem"); // NOI18N
        fileMenu.add(exitMenuItem);

        menuBar.add(fileMenu);

        editMenu.setText(resourceMap.getString("editMenu.text")); // NOI18N
        editMenu.setName("editMenu"); // NOI18N

        undoMenuItem.setAction(actionMap.get("undoAction")); // NOI18N
        undoMenuItem.setName("undoMenuItem"); // NOI18N
        editMenu.add(undoMenuItem);

        redoMenuItem.setAction(actionMap.get("redoAction")); // NOI18N
        redoMenuItem.setName("redoMenuItem"); // NOI18N
        editMenu.add(redoMenuItem);

        editSeparator1.setName("editSeparator1"); // NOI18N
        editMenu.add(editSeparator1);

        cutMenuItem.setAction(actionMap.get("cut"));
        cutMenuItem.setName("cutMenuItem"); // NOI18N
        editMenu.add(cutMenuItem);

        copyMenuItem.setAction(actionMap.get("copy"));
        copyMenuItem.setName("copyMenuItem"); // NOI18N
        editMenu.add(copyMenuItem);

        pasteMenuItem.setAction(actionMap.get("paste"));
        pasteMenuItem.setName("pasteMenuItem"); // NOI18N
        editMenu.add(pasteMenuItem);

        menuBar.add(editMenu);

        debugMenu.setText(resourceMap.getString("debugMenu.text")); // NOI18N
        debugMenu.setName("debugMenu"); // NOI18N

        runMenuItem.setAction(actionMap.get("runAction")); // NOI18N
        runMenuItem.setName("runMenuItem"); // NOI18N
        debugMenu.add(runMenuItem);

        stopMenuItem.setAction(actionMap.get("stopAction")); // NOI18N
        stopMenuItem.setName("stopMenuItem"); // NOI18N
        debugMenu.add(stopMenuItem);

        pauseMenuItem.setAction(actionMap.get("pauseAction")); // NOI18N
        pauseMenuItem.setName("pauseMenuItem"); // NOI18N
        debugMenu.add(pauseMenuItem);

        continueMenuItem.setAction(actionMap.get("continueAction")); // NOI18N
        continueMenuItem.setName("continueMenuItem"); // NOI18N
        debugMenu.add(continueMenuItem);

        debugSeparator1.setName("debugSeparator1"); // NOI18N
        debugMenu.add(debugSeparator1);

        toggleBreakpointMenuItem.setAction(actionMap.get("toggleBreakpointAction")); // NOI18N
        toggleBreakpointMenuItem.setName("toggleBreakpointMenuItem"); // NOI18N
        debugMenu.add(toggleBreakpointMenuItem);

        debugSeparator2.setName("debugSeparator2"); // NOI18N
        debugMenu.add(debugSeparator2);

        stepOverMenuItem.setAction(actionMap.get("stepOverAction")); // NOI18N
        stepOverMenuItem.setName("stepOverMenuItem"); // NOI18N
        debugMenu.add(stepOverMenuItem);

        stepIntoMenuItem.setAction(actionMap.get("stepIntoAction")); // NOI18N
        stepIntoMenuItem.setName("stepIntoMenuItem"); // NOI18N
        debugMenu.add(stepIntoMenuItem);

        menuBar.add(debugMenu);

        settingsMenu.setText(resourceMap.getString("settingsMenu.text")); // NOI18N
        settingsMenu.setName("settingsMenu"); // NOI18N

        animationSpeedMenu.setText(resourceMap.getString("animationSpeedMenu.text")); // NOI18N
        animationSpeedMenu.setName("animationSpeedMenu"); // NOI18N

        asDisabledMenuItem.setAction(actionMap.get("changeAnimationSpeed")); // NOI18N
        asDisabledMenuItem.setAccelerator(resourceMap.getKeyStroke("asDisabledMenuItem.accelerator"));
        animSpeedGroup.add(asDisabledMenuItem);
        asDisabledMenuItem.setText(resourceMap.getString("asDisabledMenuItem.text")); // NOI18N
        asDisabledMenuItem.setActionCommand(resourceMap.getString("asDisabledMenuItem.actionCommand")); // NOI18N
        asDisabledMenuItem.setName("asDisabledMenuItem"); // NOI18N
        animationSpeedMenu.add(asDisabledMenuItem);

        asSlowMenuItem.setAction(actionMap.get("changeAnimationSpeed")); // NOI18N
        asSlowMenuItem.setAccelerator(resourceMap.getKeyStroke("asSlowMenuItem.accelerator"));
        animSpeedGroup.add(asSlowMenuItem);
        asSlowMenuItem.setText(resourceMap.getString("asSlowMenuItem.text")); // NOI18N
        asSlowMenuItem.setActionCommand(resourceMap.getString("asSlowMenuItem.actionCommand")); // NOI18N
        asSlowMenuItem.setName("asSlowMenuItem"); // NOI18N
        animationSpeedMenu.add(asSlowMenuItem);

        asNormalMenuItem.setAction(actionMap.get("changeAnimationSpeed")); // NOI18N
        asNormalMenuItem.setAccelerator(resourceMap.getKeyStroke("asNormalMenuItem.accelerator"));
        animSpeedGroup.add(asNormalMenuItem);
        asNormalMenuItem.setText(resourceMap.getString("asNormalMenuItem.text")); // NOI18N
        asNormalMenuItem.setActionCommand(resourceMap.getString("asNormalMenuItem.actionCommand")); // NOI18N
        asNormalMenuItem.setName("asNormalMenuItem"); // NOI18N
        animationSpeedMenu.add(asNormalMenuItem);

        asFastMenuItem.setAction(actionMap.get("changeAnimationSpeed")); // NOI18N
        asFastMenuItem.setAccelerator(resourceMap.getKeyStroke("asFastMenuItem.accelerator"));
        animSpeedGroup.add(asFastMenuItem);
        asFastMenuItem.setText(resourceMap.getString("asFastMenuItem.text")); // NOI18N
        asFastMenuItem.setActionCommand(resourceMap.getString("asFastMenuItem.actionCommand")); // NOI18N
        asFastMenuItem.setName("asFastMenuItem"); // NOI18N
        animationSpeedMenu.add(asFastMenuItem);

        asVFastMenuItem.setAction(actionMap.get("changeAnimationSpeed")); // NOI18N
        asVFastMenuItem.setAccelerator(resourceMap.getKeyStroke("asVFastMenuItem.accelerator"));
        animSpeedGroup.add(asVFastMenuItem);
        asVFastMenuItem.setSelected(true);
        asVFastMenuItem.setText(resourceMap.getString("asVFastMenuItem.text")); // NOI18N
        asVFastMenuItem.setActionCommand(resourceMap.getString("asVFastMenuItem.actionCommand")); // NOI18N
        asVFastMenuItem.setName("asVFastMenuItem"); // NOI18N
        animationSpeedMenu.add(asVFastMenuItem);

        settingsMenu.add(animationSpeedMenu);

        menuBar.add(settingsMenu);

        windowMenu.setText(resourceMap.getString("windowMenu.text")); // NOI18N
        windowMenu.setName("windowMenu"); // NOI18N

        showOutputMenuItem.setAction(actionMap.get("showOutputAction")); // NOI18N
        showOutputMenuItem.setName("showOutputMenuItem"); // NOI18N
        windowMenu.add(showOutputMenuItem);

        showGEMenuItem.setAction(actionMap.get("showGEAction")); // NOI18N
        showGEMenuItem.setName("showGEMenuItem"); // NOI18N
        windowMenu.add(showGEMenuItem);

        showWatchesMenuItem.setAction(actionMap.get("showWatchesAction")); // NOI18N
        showWatchesMenuItem.setName("showWatchesMenuItem"); // NOI18N
        windowMenu.add(showWatchesMenuItem);

        windowSeparator1.setName("windowSeparator1"); // NOI18N
        windowMenu.add(windowSeparator1);

        clearOutputMenuItem.setAction(actionMap.get("clearOutputAction")); // NOI18N
        clearOutputMenuItem.setName("clearOutputMenuItem"); // NOI18N
        windowMenu.add(clearOutputMenuItem);

        resetGEMenuItem.setAction(actionMap.get("resetGEAction")); // NOI18N
        resetGEMenuItem.setName("resetGEMenuItem"); // NOI18N
        windowMenu.add(resetGEMenuItem);

        menuBar.add(windowMenu);

        statusPanel.setName("statusPanel"); // NOI18N
        statusPanel.setPreferredSize(new java.awt.Dimension(400, 20));
        statusPanel.setRequestFocusEnabled(false);

        statusMessageLabel.setFont(resourceMap.getFont("statusMessageLabel.font")); // NOI18N
        statusMessageLabel.setName("statusMessageLabel"); // NOI18N

        statusLineNumber.setFont(resourceMap.getFont("statusLineNumber.font")); // NOI18N
        statusLineNumber.setName("statusLineNumber"); // NOI18N

        javax.swing.GroupLayout statusPanelLayout = new javax.swing.GroupLayout(statusPanel);
        statusPanel.setLayout(statusPanelLayout);
        statusPanelLayout.setHorizontalGroup(
            statusPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(javax.swing.GroupLayout.Alignment.TRAILING, statusPanelLayout.createSequentialGroup()
                .addComponent(statusMessageLabel, javax.swing.GroupLayout.DEFAULT_SIZE, 169, Short.MAX_VALUE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.UNRELATED)
                .addComponent(statusLineNumber, javax.swing.GroupLayout.PREFERRED_SIZE, 211, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addContainerGap())
        );
        statusPanelLayout.setVerticalGroup(
            statusPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(statusLineNumber, javax.swing.GroupLayout.Alignment.TRAILING, javax.swing.GroupLayout.DEFAULT_SIZE, 20, Short.MAX_VALUE)
            .addComponent(statusMessageLabel, javax.swing.GroupLayout.DEFAULT_SIZE, 20, Short.MAX_VALUE)
        );

        setComponent(mainPanel);
        setMenuBar(menuBar);
        setStatusBar(statusPanel);
    }// </editor-fold>//GEN-END:initComponents

    javax.swing.Action getAction(String actionName) {
        return getContext().getActionMap(getClass(), this).get(actionName);
    }

    private void showErrorDialog(String message, Exception e) {
        String title = "Error";
        int type = JOptionPane.ERROR_MESSAGE;
        message = "Error: " + message;
        JOptionPane.showMessageDialog(getFrame(), message, title, type);
    }

    /**
     *  and synchronizes title bar with it. Skips
     * sequences of settings to true. Always sets isCompiled to false.
     * <p>
     * More detailed, old and new state combinations:<p>
     * old=false, new=true: char inserted when not modified, update title<br>
     * old=false, new=false: New/Open/Save fired, update title<br>
     * old=true, new=true: char inserted when already modified, skip<br>
     * old=true, new=false: New/Open/Save fired, update title<br>
     *
     * @param newModified new value of modified field
     */
    private void updateModified(boolean newModified) {
        if (!(isModified && newModified)) {
            String mod = newModified ? "*" : "";
            String fileName = (file == null ? "untitled" : file.getName());
            getFrame().setTitle(mod + fileName + " - " + appTitle);
            isModified = newModified;
        }

        isCompiled = false;
    }

    public boolean isRunning() {
        return running;
    }

    public boolean isNotRunning() {
        return !running;
    }

    private void setRunning(boolean running) {
        boolean oldValue = this.running;
        this.running = running;
        firePropertyChange(RUNNING_PROPERTY, oldValue, running);
        firePropertyChange(NOT_RUNNING_PROPERTY, !oldValue, !running);
    }

    public boolean isPaused() {
        return paused;
    }

    private void setPaused(boolean paused) {
        boolean oldValue = this.paused;
        this.paused = paused;
        firePropertyChange(PAUSED_PROPERTY, oldValue, paused);
    }

    public boolean isStopped() {
        return stopped;
    }

    public boolean isNotStopped() {
        return !stopped;
    }

    private void setStopped(boolean stopped) {
        boolean oldValue = this.stopped;
        this.stopped = stopped;
        firePropertyChange(STOPPED_PROPERTY, oldValue, stopped);
        firePropertyChange(NOT_STOPPED_PROPERTY, !oldValue, !stopped);
    }

    @Action(enabledProperty = STOPPED_PROPERTY)
    public void newAction() {
        if (canCloseFile()) {
            clearBreakpoints();
            fingyPane.getTextPanel().setText("");
            file = null;
            undo.discardAllEdits();
            updateModified(false);
        }
    }

    @Action(enabledProperty = STOPPED_PROPERTY)
    public void openAction() {
        if (canCloseFile()) {
            int option = fingyScriptFileChooser.showOpenDialog(getFrame());
            if (option == JFileChooser.APPROVE_OPTION) {
                file = fingyScriptFileChooser.getSelectedFile();
                try {
                    FileInputStream fis = new FileInputStream(file);
                    clearBreakpoints();
                    fingyPane.getTextPanel().setText("");
                    fingyPane.getTextPanel().read(fis, null);
                    undo.discardAllEdits();
                    updateModified(false);
                    statusMessageLabel.setText("File " + file.getName() + " opened");
                } catch (IOException e) {
                    showErrorDialog("could not open file \"" + file + "\"", e);
                }
            }
        }
    }

    private boolean doSave() {
        boolean saved = false;
        try {
            fingyPane.getTextPanel().write(new FileWriter(file));
            saved = true;
        } catch (IOException e) {
            showErrorDialog("could not write file \"" + file + "\"", e);
        }

        return saved;
    }

    private boolean doSaveAs() {
        boolean saved = false;
        int option = fingyScriptFileChooser.showSaveDialog(getFrame());
        if (option == JFileChooser.APPROVE_OPTION) {
            file = fingyScriptFileChooser.getSelectedFile();
            try {
                fingyPane.getTextPanel().write(new FileWriter(file));
                saved = true;
            } catch (IOException e) {
                showErrorDialog("could not write file \"" + file + "\"", e);
            }
        }

        return saved;
    }

    private boolean canCloseFile() {
        if (isModified) {
            String message = "Save file \""
                    + (file == null ? "untitled" : file.getName()) + "\" ?";
            int option = JOptionPane.showConfirmDialog(getFrame(), message,
                    "Save", JOptionPane.YES_NO_CANCEL_OPTION);

            switch (option) {
                case JOptionPane.YES_OPTION:
                    saveAction();
                    return true;
                case JOptionPane.NO_OPTION:
                    return true;
                case JOptionPane.CANCEL_OPTION:
                case JOptionPane.CLOSED_OPTION:
                    return false;
            }
        }
        return true;
    }

    @Action(enabledProperty = STOPPED_PROPERTY)
    public void saveAction() {
        boolean saved = (file == null ? doSaveAs() : doSave());

        if (saved) {
            updateModified(false);
            statusMessageLabel.setText("File " + file.getName() + " saved");
        }
    }

    @Action(enabledProperty = STOPPED_PROPERTY)
    public void saveAsAction() {
        if (doSaveAs()) {
            updateModified(false);
            statusMessageLabel.setText("Saved to " + file.getName());
        }
    }

    @Action(enabledProperty = STOPPED_PROPERTY)
    public void undoAction() {
        if (undo.canUndo()) {
            undo.undo();
            updateModified(true);
            statusMessageLabel.setText("Undo previous change");
        } else {
            statusMessageLabel.setText("Can't undo change");
        }
    }

    @Action(enabledProperty = STOPPED_PROPERTY)
    public void redoAction() {
        if (undo.canRedo()) {
            undo.redo();
            statusMessageLabel.setText("Redo change");
        } else {
            statusMessageLabel.setText("Can't redo change");
        }
    }

    private void regGE(GameEnvironment ge) {
        ge.initInterpreter(itp);
        fingyDocument.addFunctions(ge.getFunctionList());
        fingyDocument.addConstants(ge.getConstantList());
        fingyDocument.updateHighlighting();
    }

    private void unregGE(GameEnvironment ge) {
        ge.purgeInterpreter(itp);
        fingyDocument.removeFunctions(ge.getFunctionList());
        fingyDocument.removeConstants(ge.getConstantList());
        fingyDocument.updateHighlighting();
    }

    // workaround method
    // TODO: GE class name should be provided in xml
    private Class<? extends GameEnvironment> getGEClass(String name) {
        if (name.equals("Turtle")) {
            return TurtleGameEnvironment.class;
        } else if (name.equals("Robot")) {
            return RobotGameEnvironment.class;
        }

        return null;
    }

    // mostly workaround method
    // TODO: all GE default parameters (like width, height, etc) should be provided in xml and eaten by ge itself
    private GameEnvironment getGEInstance(Class<? extends GameEnvironment> geClass, final Node geNode) {
        GameEnvironment ge = null;
        if (geClass.equals(TurtleGameEnvironment.class)) {
            int turtleWidth = 500;
            int turtleHeight = 500;
            ge = new TurtleGameEnvironment(turtleWidth, turtleHeight);
        } else if (geClass.equals(RobotGameEnvironment.class)) {
            int fieldWidth = 10;
            int fieldHeight = 10;
            int width = (fieldWidth + 1) * 55;
            int height = (fieldHeight + 2) * 55;
            ge = new RobotGameEnvironment(width, height, fieldWidth, fieldHeight);
        }

        return ge;
    }

    private void initGEWindow() {
        gePanel = new GameEnvPanel();
        geWindow = createPanelWindow("Game Environment", gePanel);
    }

    private void reloadGEWindow(String title, boolean resizable) {
        gePanel.setComponent(gameEnv.getVisualComponent());
        geWindow.setTitle(title);
        geWindow.setResizable(resizable);
        geWindow.pack();
    }

    @Action(enabledProperty = STOPPED_PROPERTY)
    public void loadGEAction() {
        int option = gameEnvironmentFileChooser.showOpenDialog(getFrame());
        if (option == JFileChooser.APPROVE_OPTION) {

            /* plan (remove when all is known):
             *      read xml
             *      if doc is ok, go on
             *
             *      clear breaks
             *      unreg current GE
             *          unreg GE functions in Interpreter
             *          unreg GE functions in FingyDocument
             *
             *      decide which GE class to load
             *          no className workaround()
             *      no default parsing by GE workaround()
             *          get default params
             *          drop old ge
             *          create instance of that class with default params
             *
             *      reg new GE
             *          init Interpeter w/ functions of new GE
             *          add functions of new GE to FingyDocument
             *      reset new GE
             *      refresh FingyDocument?
             *
             *      replace GameEnvPanel's main component with visual component of new GE
             */

            File f = gameEnvironmentFileChooser.getSelectedFile();
            try {
                // try read XML
                DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                DocumentBuilder db = dbf.newDocumentBuilder();
                Document doc = db.parse(f);
                doc.getDocumentElement().normalize();
                if (!doc.getDocumentElement().getNodeName().equals("FingyApplicationConfig")) {
                    System.err.println("Corrupted or incorrect config file.");
                    statusMessageLabel.setText("Incorrect or inaccessible xml file");
                    return;
                }

                // XML is ok, clear breaks, unreg current GE
                // clearBreakpoints();

                // TODO: make one single property for checking if GE is loaded, instead of 3 nullchecks
                if (gameEnv != null) {
                    unregGE(gameEnv);
                    isCompiled = false;
                }

                // decide which GE class to load: get it's name and params
                NodeList geNodes = doc.getElementsByTagName("GameEnvironment");
                final Node geNode = geNodes.item(0);
                String geName = geNode.getAttributes().getNamedItem("name").getNodeValue();

                // drop previous GE (or get OutOfMemory and die)
                if (gePanel != null) {
                    gePanel.setComponent(null);
                }
                gameEnv = null;
                setGEWindowLoaded(false);

                // load new GE, passing it it's params
                gameEnv = getGEInstance(getGEClass(geName), geNode);

                regGE(gameEnv);
                resetGEAction();

                new Thread() {

                    @Override
                    public void run() {
                        gameEnv.loadFromXML(geNode);
                    }
                }.start();

                if (!isGEWindowLoaded()) {
                    initGEWindow();
                    arrangeGEWindow(geWindow);
                }
                reloadGEWindow(geName, true);
                setGEWindowLoaded(true);
                changeAnimationSpeed();

                statusMessageLabel.setText("GE state loaded");
            } catch (Exception e) {
                e.printStackTrace();
                statusMessageLabel.setText("Incorrect or inaccessible xml file");
            }
        }
    }

    /*
     * Recompiles the source code, if necessary.
     */
    private void recompile() {
        outputPanel.clear();
        if (!isCompiled) {
            try {
                itp.compile(fingyPane.getText());
                isCompiled = true;
            } catch (SyntaxError ex) {
                outputPanel.printfError("Syntax error on line %1d: %2s\n",
                        ex.getLineNumber(), ex.getMessage());
            }
        }
    }

    /*
     * Creates new VMContext from currently compiled code and loads it in
     * RunServer.
     */
    private void restartRunServer() {
        VMContext ctx = itp.createVMContext();
        ctx.setDebug(true);
        runner.restart(ctx);
        runner.addBreakpoints(breakpoints);
    }

    @Action(enabledProperty = STOPPED_PROPERTY)
    public void runAction() {
        if (running) {
            return;
        }
        statusMessageLabel.setText("Running...");
        recompile();

        // don't run if syntax error occured
        if (isCompiled) {
            restartRunServer();
            runner.run();
        }
    }

    @Action(enabledProperty = NOT_STOPPED_PROPERTY)
    public void stopAction() {
        statusMessageLabel.setText("Stopped");
        runner.stop();
    }

    @Action(enabledProperty = RUNNING_PROPERTY)
    public void pauseAction() {
        statusMessageLabel.setText("Paused");
        runner.pause();
    }

    @Action(enabledProperty = PAUSED_PROPERTY)
    public void continueAction() {
        if (!runner.isFinished()) {
            statusMessageLabel.setText("Continuing...");
            debugOutput("continuing from line " + runner.getCurrentLine());
            runner.syncBreakpoints(breakpoints);
            runner.run();
        }
    }

    private Set<Integer> breakpoints = new HashSet<Integer>();

    @Action(enabledProperty = NOT_RUNNING_PROPERTY)
    public void toggleBreakpointAction() {

        int line = fingyPane.getLine() + 1;

        // new state of breakpoint at current line
        boolean enabled = !breakpoints.contains(line);

        if (enabled) {
            breakpoints.add(line);
            statusMessageLabel.setText("Breakpoint added at line " + line);
        } else {
            breakpoints.remove(line);
            statusMessageLabel.setText("Breakpoint removed at line " + line);
        }

        // FingyDocument counts lines from 0, IDE - from 1
        fingyDocument.setBreakLine(line - 1, enabled);
    }

    /*
     * Clears IDE (visible) breakpoints only. Breakpoints in RunServer
     * and it's VMContext are not affected.
     */
    private void clearBreakpoints() {
        breakpoints.clear();
        fingyDocument.clearBreakLines();
    }

    private void preStep() {
        if (runner.isFinished()) {
            // current VMContext is dead == we're starting program
            // from the beginning, using Step Over (Step Into)
            recompile();
            if (isCompiled) {
                restartRunServer();
            }
        } else {
            // current VMContext is alive == we're continuing execution
            runner.syncBreakpoints(breakpoints);
        }
    }

    @Action(enabledProperty = NOT_RUNNING_PROPERTY)
    public void stepOverAction() {
        statusMessageLabel.setText("Step Over");
        preStep();
        if (isCompiled) {
            runner.stepOver();
        }
    }

    @Action(enabledProperty = NOT_RUNNING_PROPERTY)
    public void stepIntoAction() {
        statusMessageLabel.setText("Step Into");
        preStep();
        if (isCompiled) {
            runner.stepInto();
        }
    }

    @Action(enabledProperty = GE_WINDOW_LOADED_PROPERTY)
    public void changeAnimationSpeed() {
        if (gameEnv == null) {
            return;
        }

        String speed = animSpeedGroup.getSelection().getActionCommand();

        if ("animation_disabled".equals(speed)) {
            gameEnv.setAnimationSpeed(AnimationSpeed.DISABLED);
        } else if ("animation_slow".equals(speed)) {
            gameEnv.setAnimationSpeed(AnimationSpeed.SLOW);
        } else if ("animation_normal".equals(speed)) {
            gameEnv.setAnimationSpeed(AnimationSpeed.NORMAL);
        } else if ("animation_fast".equals(speed)) {
            gameEnv.setAnimationSpeed(AnimationSpeed.FAST);
        } else if ("animation_vfast".equals(speed)) {
            gameEnv.setAnimationSpeed(AnimationSpeed.VERYFAST);
        }

        statusMessageLabel.setText("Animation Speed set to: " + speed.substring(10));
    }

    @Action
    public void showOutputAction() {
        statusMessageLabel.setText("Show Output");
        outputWindow.setVisible(true);
    }

    public boolean isGEWindowLoaded() {
        return geWindow != null;
    }

    private void setGEWindowLoaded(boolean loaded) {
        firePropertyChange(GE_WINDOW_LOADED_PROPERTY, !loaded, loaded);
    }

    @Action(enabledProperty = GE_WINDOW_LOADED_PROPERTY)
    public void showGEAction() {
        statusMessageLabel.setText("Show Game Environment");
        geWindow.setVisible(true);
    }

    @Action
    public void showWatchesAction() {
        statusMessageLabel.setText("Show Watches");
        watchWindow.setVisible(true);
    }

    @Action(enabledProperty = STOPPED_PROPERTY)
    public void clearOutputAction() {
        statusMessageLabel.setText("Clear Output");
        outputPanel.clear();
    }

    @Action(enabledProperty = STOPPED_PROPERTY)
    public void resetGEAction() {
        statusMessageLabel.setText("Reset GE");

        new Thread() {

            @Override
            public void run() {
                gameEnv.resetState();
            }
        }.start();
    }

    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.ButtonGroup animSpeedGroup;
    private javax.swing.JMenu animationSpeedMenu;
    private javax.swing.JRadioButtonMenuItem asDisabledMenuItem;
    private javax.swing.JRadioButtonMenuItem asFastMenuItem;
    private javax.swing.JRadioButtonMenuItem asNormalMenuItem;
    private javax.swing.JRadioButtonMenuItem asSlowMenuItem;
    private javax.swing.JRadioButtonMenuItem asVFastMenuItem;
    private javax.swing.JMenuItem clearOutputMenuItem;
    private javax.swing.JMenuItem continueMenuItem;
    private javax.swing.JMenuItem copyMenuItem;
    private javax.swing.JMenuItem cutMenuItem;
    private javax.swing.JMenu debugMenu;
    private javax.swing.JPopupMenu.Separator debugSeparator1;
    private javax.swing.JPopupMenu.Separator debugSeparator2;
    private javax.swing.JMenu editMenu;
    private javax.swing.JPopupMenu.Separator editSeparator1;
    private javax.swing.JPopupMenu.Separator fileSeparator1;
    private javax.swing.JMenuItem loadGEMenuItem;
    private javax.swing.JPanel mainPanel;
    private javax.swing.JMenuBar menuBar;
    private javax.swing.JMenuItem newMenuItem;
    private javax.swing.JMenuItem openMenuItem;
    private javax.swing.JMenuItem pasteMenuItem;
    private javax.swing.JMenuItem pauseMenuItem;
    private javax.swing.JMenuItem redoMenuItem;
    private javax.swing.JMenuItem resetGEMenuItem;
    private javax.swing.JMenuItem runMenuItem;
    private javax.swing.JMenuItem saveAsMenuItem;
    private javax.swing.JMenuItem saveMenuItem;
    private javax.swing.JMenu settingsMenu;
    private javax.swing.JMenuItem showGEMenuItem;
    private javax.swing.JMenuItem showOutputMenuItem;
    private javax.swing.JMenuItem showWatchesMenuItem;
    private javax.swing.JPanel sourceEditorPanel;
    private javax.swing.JLabel statusLineNumber;
    private javax.swing.JLabel statusMessageLabel;
    private javax.swing.JPanel statusPanel;
    private javax.swing.JMenuItem stepIntoMenuItem;
    private javax.swing.JMenuItem stepOverMenuItem;
    private javax.swing.JMenuItem stopMenuItem;
    private javax.swing.JMenuItem toggleBreakpointMenuItem;
    private javax.swing.JMenuItem undoMenuItem;
    private javax.swing.JMenu windowMenu;
    private javax.swing.JPopupMenu.Separator windowSeparator1;
    // End of variables declaration//GEN-END:variables
    private FingySourceEditor fingyPane;
    private JFileChooser fingyScriptFileChooser;
    private JFileChooser gameEnvironmentFileChooser;
    private String appTitle;
    private File file;
    private JDialog outputWindow;
    private JDialog watchWindow;
    private JDialog geWindow;
    private OutputPanel outputPanel;
    private WatchPanel watchPanel;
    private GameEnvPanel gePanel;
    private Interpreter itp;
    private RunServer runner;
    private GameEnvironment gameEnv;
    private boolean isModified = false;
    private boolean isCompiled = false;
    private boolean running = false;
    private boolean stopped = true;
    private boolean paused = false;
    private boolean spamDebug = false;
}
