/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010. Alexey Rusakovich <mailto:exp_sm@yahoo.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.ide;

import by.informatics.jcx.fingy.lang.BoolValue;
import by.informatics.jcx.fingy.lang.FloatValue;
import by.informatics.jcx.fingy.lang.IntValue;
import by.informatics.jcx.fingy.lang.NullValue;
import by.informatics.jcx.fingy.lang.StringValue;
import by.informatics.jcx.fingy.lang.Value;
import by.informatics.jcx.fingy.lang.ValueType;
import java.awt.Color;
import java.awt.Frame;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import javax.swing.ButtonGroup;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JRadioButton;
import javax.swing.JTextField;

abstract class InputDialog extends JDialog implements PropertyChangeListener {

    private static final long serialVersionUID = 1L;

    protected static final String title = "Input";
    protected static final String enterButton = "Enter";
    protected static final Object[] buttons = { enterButton, "Stop" };
    protected JOptionPane optionPane;
    protected Object[] defaultDisplayItems;

    protected Value returnValue = NullValue.NULL;

    public static InputDialog getDialog(Frame owner, ValueType type) {
        switch (type) {
            case BOOL:
                return new BoolInputDialog(owner);
            case INT:
                return new IntInputDialog(owner);
            case FLOAT:
                return new FloatInputDialog(owner);
            case STRING:
                return new StringInputDialog(owner);
        }

        return null;
    }

    public InputDialog(Frame owner) {
        super(owner, title, true);
    }

    protected void finishInit(Object[] displayItems) {
        defaultDisplayItems = displayItems;

        optionPane = new JOptionPane(displayItems,
                JOptionPane.QUESTION_MESSAGE,
                JOptionPane.YES_NO_OPTION,
                null,
                buttons,
                enterButton);

        // make this dialog display it
        setContentPane(optionPane);

        setResizable(false);
        pack();

        // handle window closing correctly
        setDefaultCloseOperation(DO_NOTHING_ON_CLOSE);
        addWindowListener(new WindowAdapter() {

            @Override
            public void windowClosing(WindowEvent e) {
                /*
                 * Instead of directly closing the window,
                 * we're changing the JOptionPane's value property.
                 */
                optionPane.setValue(JOptionPane.CLOSED_OPTION);
            }
        });

        // register an event handler that reacts to option pane state changes
        optionPane.addPropertyChangeListener(this);
    }

    protected abstract void processValueChange();

    public void propertyChange(PropertyChangeEvent evt) {
        String prop = evt.getPropertyName();

        boolean isOurEvent = isVisible()
                && (evt.getSource() == optionPane)
                && (JOptionPane.VALUE_PROPERTY.equals(prop)
                || JOptionPane.INPUT_VALUE_PROPERTY.equals(prop));

        if (!isOurEvent) {
            return;  // this is not our event
        }

        Object value = optionPane.getValue();

        if (value == JOptionPane.UNINITIALIZED_VALUE) {
            return;  // ignore reset
        }

        // Reset the JOptionPane's value.
        // If we don't, then if the user presses the same button next time,
        // no property change event will be fired.
        optionPane.setValue(JOptionPane.UNINITIALIZED_VALUE);

        if (enterButton.equals(value)) {
            processValueChange();
        } else {
            // user closed dialog or clicked cancel
            System.out.println("[debug] InputDialog.propertyChange(): Stopped by user");
            returnValue = NullValue.NULL;
            setVisible(false);
        }
    }

    protected void reset() {
    }

    public Value getValue(String message) {
        reset();
        setLocationRelativeTo(getOwner());
        if (message != null) {
            Object[] o = new Object[3];
            o[0] = message;
            o[1] = ((Object[])optionPane.getMessage())[1];
            o[2] = ((Object[])optionPane.getMessage())[2];
            optionPane.setMessage(o);
        }
        else {
            optionPane.setMessage(defaultDisplayItems);
        }
        setVisible(true);
        return returnValue;
    }
}

class BoolInputDialog extends InputDialog {
    private static final long serialVersionUID = 1L;
    private static final String message = "Enter bool:";
    private static final String trueCommand = "true";
    private static final String falseCommand = "false";
    private final ButtonGroup group = new ButtonGroup();
    private final JRadioButton btnTrue = new JRadioButton("TRUE", true);
    private final JRadioButton btnFalse = new JRadioButton("FALSE");

    public BoolInputDialog(Frame owner) {
        super(owner);

        btnTrue.setActionCommand(trueCommand);
        btnFalse.setActionCommand(falseCommand);
        group.add(btnTrue);
        group.add(btnFalse);

        Object[] displayItems = { message, btnTrue, btnFalse };
        finishInit(displayItems);
    }

    @Override
    protected void processValueChange() {
        String cmd = group.getSelection().getActionCommand();
        returnValue = BoolValue.create(Boolean.parseBoolean(cmd));
        setVisible(false);
    }

    @Override
    protected void reset() {
        btnTrue.setSelected(true);
    }
}

abstract class TextInputDialog extends InputDialog {

    private static final long serialVersionUID = 1L;

    protected final JTextField textField = new JTextField();

    public TextInputDialog(Frame owner) {
        super(owner);

        addComponentListener(new ComponentAdapter() {

            @Override
            public void componentShown(ComponentEvent ce) {
                textField.requestFocusInWindow();
            }
        });
    }

    @Override
    protected void reset() {
        textField.setText(null);
    }
}

abstract class ValidatedInputDialog extends TextInputDialog {

    private static final long serialVersionUID = 1L;

    protected final JLabel errorLabel = new JLabel(" ");

    public ValidatedInputDialog(Frame owner, String message) {
        super(owner);

        errorLabel.setForeground(Color.RED);

        Object[] displayItems = { message, textField, errorLabel };
        finishInit(displayItems);
    }

    protected void processError() {
        errorLabel.setText("Invalid input.");
        textField.selectAll();
        textField.requestFocusInWindow();
    }

    @Override
    protected void reset() {
        super.reset();
        errorLabel.setText(" ");
    }
}

class IntInputDialog extends ValidatedInputDialog {
    private static final long serialVersionUID = 1L;
    private static final String message = "Enter int:";

    public IntInputDialog(Frame owner) {
        super(owner, message);
    }

    @Override
    protected void processValueChange() {
        int x;
        try {
            x = Integer.parseInt(textField.getText());
            returnValue = IntValue.create(x);
            setVisible(false);
        } catch (NumberFormatException ex) {
            processError();
        }
    }
}

class FloatInputDialog extends ValidatedInputDialog {
    private static final long serialVersionUID = 1L;
    private static final String message = "Enter float:";

    public FloatInputDialog(Frame owner) {
        super(owner, message);
    }

    @Override
    protected void processValueChange() {
        double f;
        try {
            f = Double.parseDouble(textField.getText());
            returnValue = FloatValue.create(f);
            setVisible(false);
        } catch (NumberFormatException ex) {
            processError();
        }
    }
}

class StringInputDialog extends TextInputDialog {
    private static final long serialVersionUID = 1L;
    private static final String message = "Enter string:";

    public StringInputDialog(Frame owner) {
        super(owner);

        Object[] displayItems = {message, textField};
        finishInit(displayItems);
    }

    @Override
    protected void processValueChange() {
        returnValue = StringValue.create(textField.getText());
        setVisible(false);
    }
}
