/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010. Alexey Rusakovich <mailto:exp_sm@yahoo.com>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.ide;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.SwingUtilities;

import by.informatics.jcx.fingy.lang.Breakpoint;
import by.informatics.jcx.fingy.lang.LangException;
import by.informatics.jcx.fingy.lang.VMContext;
import by.informatics.jcx.fingy.lang.VMTerminateSignal;
import by.informatics.jcx.fingy.lang.Variable;

public class RunServer {

    private class RunThread extends Thread {

        private void processError(final LangException e) {
            SwingUtilities.invokeLater(new Runnable() {

                @Override
                public void run() {
                    client.processRuntimeError(e);
                }
            });
        }

        private boolean stopRequired = false;

        @Override
        public void run() {
            client.debugOutput("Server: started, switching to wait mode...");

            while (!stopRequired) {
                client.debugOutput("Server: iteration");
                try {
                    client.debugOutput("Server: pre-sync");
                    synchronized (this) {
                        client.debugOutput("Server: post-sync, pre-wait...");
                        while (cmd ==  null && !stopRequired) {
                            client.debugOutput("Server: waiting...");
                            wait();
                        }
                        client.debugOutput("Server: wait OK! cmd = " + cmd + ", stopRequired = " + stopRequired);
                    }
                } catch (InterruptedException e) {
                }

                if (stopRequired) {
                    finished = true;
                    client.processStop(true);
                    break;
                }

                try {
                    client.processRun();
                    switch (cmd) {
                        case RUN:
                            finished = !context.run();
                            break;
                        case STEPOVER:
                            finished = !context.runLine(false);
                            break;
                        case STEPINTO:
                            finished = !context.runLine(true);
                            break;
                    }
                    cmd = null;  // consume command
                } catch (VMTerminateSignal ex) {
                    finished = true;
                    client.debugOutput("caught VMTerminalSignal in RunThread, stopping...");
                } catch (LangException ex) {
                    finished = true;
                    processError(ex);
                } finally {
                    stopRequired = finished;
                    client.processStop(finished);

                    if (!finished) {
                        client.debugOutput("Server: firing pause..");
                    } else {
                        client.debugOutput("Server: finished execution");
                    }
                }
            }

            client.debugOutput("Server: all jobs done, closing server...");
        }

        synchronized void go() {
            notify();
            client.debugOutput("RunServer: post-notify");
        }

        synchronized void requireStop() {
            stopRequired = true;
            notify();
            client.debugOutput("RunServer: stop required");
        }
    }

    private final FingyView client;
    private VMContext context;
    private RunThread srv;

    private boolean finished = true;

    public RunServer(FingyView view) {
        this.client = view;
    }

    /*
     * Stops previous execution thread, sets given VMContext as current
     * and starts new execution thread.
     */
    public void restart(VMContext context) {
        client.debugOutput("RunServer: restarting server core...");

        if (srv != null) {
            srv.requireStop();
        }

        if (context == null) {
            throw new IllegalArgumentException("context must be not null");
        }

        this.context = context;
        finished = false;

        srv = new RunThread();
        srv.start();
    }

    private Command cmd = null;

    public void run() {
        client.debugOutput("RunServer: run invoked");
        cmd = Command.RUN;
        srv.go();
    }

    public void stepOver() {
        client.debugOutput("RunServer: stepover invoked");
        cmd = Command.STEPOVER;
        srv.go();
    }

    public void stepInto() {
        client.debugOutput("RunServer: stepinto invoked");
        cmd = Command.STEPINTO;
        srv.go();
    }

    /*
     * Stops RunServer. Server must be restarted after it was stopped.
     */
    public void stop() {
        if (context.isRunning()) {
            context.stop();
        }
        srv.requireStop();

        context = null;
        srv = null;
    }

    public void pause() {
        context.pause();
    }

    /*
     * Returns true if the current VMContext is dead (finished).
     */
    public boolean isFinished() {
        return finished;
    }

    public int getCurrentLine() {
        return context.getCurrentLine();
    }

    public List<Collection<Variable>> getVariables() {
        if (context == null) {
            return Collections.emptyList();
        }
        return context.getVariables();
    }

    private Map<Integer, Breakpoint> vmBreaks =
            new HashMap<Integer, Breakpoint>();

    /*
     * Adds breakpoints for all lines to VMContext.
     */
    public void addBreakpoints(Set<Integer> lines) {
        for (int line : lines) {
            addBreakpoint(line);
        }
    }

    /*
     * Synchronizes current breakpoints in VMContext with given ones.
     * Missing breakpoints are added, excessive removed.
     */
    public void syncBreakpoints(Set<Integer> ideBreaks) {
        Set<Integer> vmCurBreaks = vmBreaks.keySet();

        Set<Integer> added = new HashSet<Integer>(ideBreaks);
        added.removeAll(vmCurBreaks);
        Set<Integer> removed = new HashSet<Integer>(vmCurBreaks);
        removed.removeAll(ideBreaks);

        for (int line : removed) {
            removeBreakpoint(line);
        }

        for (int line : added) {
            addBreakpoint(line);
        }
    }

    private void addBreakpoint(int line) {
        Breakpoint bp = context.createBreakpoint(line);
        vmBreaks.put(line, bp);
    }

    private void removeBreakpoint(int line) {
        vmBreaks.get(line).remove();
        vmBreaks.remove(line);
    }
}

enum Command {
    RUN,
    STEPOVER,
    STEPINTO;
}