/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Zhlobich Andrei <mailto:anjensan@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.lang;

import by.informatics.jcx.fingy.lang.modules.BuiltinModule;
import by.informatics.jcx.fingy.lang.parser.ASTNode;
import by.informatics.jcx.fingy.lang.parser.FingyParser;
import by.informatics.jcx.fingy.lang.parser.ParseException;
import by.informatics.jcx.fingy.lang.parser.TokenMgrError;
import by.informatics.jcx.fingy.lang.vm.VMCode;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.util.HashMap;
import java.util.Map;

public class Interpreter {

    private VMCode code;
    private Map<String, Function> functions = new HashMap<String, Function>();
    private Map<String, Value> values = new HashMap<String, Value>();
    private Map<String, Module> modules = new HashMap<String, Module>();

    public Interpreter() {
    }

    public void compile(String program) throws SyntaxError {
        compile(new StringReader(program));
    }

    public void compile(Reader program) throws SyntaxError {
        FingyParser parser = new FingyParser(program);
        ASTNode ast = null;
        try {
            ast = parser.parse();
        } catch (ParseException ex) {
            throw new SyntaxError(ex.currentToken.beginLine, ex.getMessage());
        } catch (TokenMgrError ex) {
            throw LexicalError.fromTokenMgrError(ex);
        }
        addModule("$usermodule", new SimpleModule(functions, values));
        addModule("$builtin", new BuiltinModule());
        code = VMCode.create(ast, modules, true);
        removeModule("");
    }

    public Function getFunction(String name) {
        return functions.get(name);
    }

    public void addFunction(String name, Function func) {
        functions.put(name, func);
    }

    public void removeFunction(String name) {
        functions.remove(name);
    }

    public void addValue(String name, Value value) {
        values.put(name, value);
    }

    public void removeValue(String name) {
        values.remove(name);
    }

    public VMContext createVMContext() {
        return new VMContext(this, code);
    }

    public void addModule(String name, Module module) {
        modules.put(name, module);
    }

    public void removeModule(String name) {
        modules.remove(name);
    }

    public static void main(String[] args) {
        Interpreter ip = new Interpreter();
        ip.addFunction("print", new Function() {

            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                for (int i = 0; i < args.size(); ++i) {
                    System.out.print(args.get(i));
                    System.out.print(" ");
                }
                System.out.println();
                return NullValue.NULL;
            }
        });

        try {
            ip.compile(new InputStreamReader(System.in));
        } catch (SyntaxError ex) {
            System.err.printf("Syntax error on line %1d: %2s\n",
                    ex.getLineNumber(), ex.getMessage());
        }
        try {
            ip.createVMContext().run();
        } catch (LangException ex) {
            System.err.printf("Runtime error on line %1d: %2s\n",
                    ex.getLineNumber(), ex.getMessage());
        }
    }
}
