/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Zhlobich Andrei <mailto:anjensan@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.lang;

import by.informatics.jcx.fingy.lang.vm.VMCode;
import by.informatics.jcx.fingy.lang.vm.VMMachine;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class VMContext {

    private VMMachine vm;
    private Thread vmThread;
    private Map<Integer, Breakpoint> breakpoints
            = new HashMap<Integer, Breakpoint>();
    private boolean needBPTracing;
    private boolean debug;

    VMContext(Interpreter itp, VMCode code) {
        vm = new VMMachine(this, code);
    }

    public boolean run() throws RuntimeError {
        preRun();
        if (needBPTracing) {
            do {
                if (!runOneCommand()) {
                    return false;
                }
            } while (!isBreakpoint());
            return true;
        } else {
            return vm.run();
        }
    }

    public boolean runLine(boolean stepInto) throws RuntimeError {
        preRun();
        if (!stepInto) {
            int d = vm.getStackDeep();
            do {
                if (!runOneCommand()) {
                    return false;
                }
            } while (!(isBreakpoint() || vm.isPaused())
                    && d < vm.getStackDeep());
            return true;
        } else {
            return runOneCommand();
        }
    }

    public void pause() {
        if (!isRunning()) {
            throw new java.lang.IllegalStateException("VM is not running");
        }
        vm.pause();
    }

    public void setDebug(boolean debug) {
        this.debug = debug;
    }

    @SuppressWarnings("deprecation")
    public void stop() {
        if (vmThread != null && !vmThread.isInterrupted()) {
            vmThread.stop(new VMTerminateSignal());
        }
    }

    public boolean isRunning() {
        return vm.isRunning();
    }

    public int getCurrentLine() {
        // line == 0 before any code interpreted
        int line = vm.getLine();
        if (line == 0) {
            line = 1;
        }
        return line;
    }

    public List<Collection<Variable>> getVariables() {
        if (!debug) {
            throw new IllegalStateException("VM is not in debug mode");
        }
        try {
            return vm.getVariables();
        } catch (RuntimeError ex) {
            throw new RuntimeException(ex);
        }
    }

    public Breakpoint createBreakpoint(int line) {
        if (breakpoints.containsKey(line)) {
            return breakpoints.get(line);
        }
        Breakpoint bp = new Breakpoint(line, this);
        addBreakpoint(bp);
        return bp;
    }

    public Breakpoint[] getBreakpoints() {
        return (Breakpoint[]) breakpoints.values().toArray();
    }

    private boolean isBreakpoint() {
        return needBPTracing && breakpoints.containsKey(getCurrentLine());
    }

    void removeBreakpoint(Breakpoint bp) {
        breakpoints.remove(bp.getLineNumber());
        needBPTracing = !breakpoints.isEmpty();
    }

    void addBreakpoint(Breakpoint bp) {
        breakpoints.put(bp.getLineNumber(), bp);
        needBPTracing = true;
    }

    private final boolean runOneCommand() throws RuntimeError {
        return vm.runLine();
    }

    private void preRun() {
        if (isRunning()) {
            throw new java.lang.IllegalStateException("VM is already running");
        }
        vmThread = Thread.currentThread();
        vm.setDebug(debug);
    }
}
