/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Zhlobich Andrei <mailto:anjensan@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.lang.modules;

import by.informatics.jcx.fingy.lang.*;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import static by.informatics.jcx.fingy.lang.ValueType.*;

public class BuiltinModule implements Module {

    public static final int MAX_ARRAY_SIZE = 9999999;

    private static Map<String, Function> functions =
            new HashMap<String, Function>();

    static {
        add("length", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeLength(args);
            }
        });

        add("array", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeArray(args);
            }
        });

        add("array2", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeArray2(args);
            }
        });

        add("to_str", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                args.checkCount(1);
                return StringValue.create(args.get(0).toString());
            }
        });

        add("to_int", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeToInt(args);
            }
        });

        add("to_float", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeToFloat(args);
            }
        });

        add("resize", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeResize(args);
            }
        });

        add("type", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeType(args);
            }
        });

        add("random", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return FloatValue.create(Math.random());
            }
        });

        add("abs", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeAbs(args);
            }
        });
    }

    @Override
    public Function getFunction(String name) {
        return functions.get(name);
    }

    @Override
    public Value getValue(String name) {
        return null;
    }

    @Override
    public String[] getFunctionNames() {
        return functions.keySet().toArray(new String[0]);
    }

    @Override
    public String[] getValueNames() {
        return null;
    }

    public static String[] getNames() {
        return functions.keySet().toArray(new String[0]);
    }

    private static void add(String name, Function function) {
        functions.put(name, function);
    }

    private static Value createArray(int size, Value def) {
        List<Value> list = new ArrayList<Value>(size);
        for (int i = 0; i < size; ++i) {
            list.add(def);
        }
        return ArrayValue.create(list);
    }

    private static Value invokeArray(VList args) throws VMError {
        args.checkMinCount(1);
        args.checkMaxCount(2);
        int size = args.get(0).getInt();
        if (size < 0) {
            throw new InvalidArgumentError("Size must be positive integer");
        }
        if (size > MAX_ARRAY_SIZE) {
            throw new InvalidArgumentError("Too big array size");
        }
        Value v = args.size() >= 2 ? args.get(1) : NullValue.NULL;
        return createArray(size, v);
    }

    private static Value invokeLength(VList args) throws VMError {
        args.checkCount(1);
        Value v = args.get(0);
        ValueType t = v.getType();
        if (t == STRING) {
            return IntValue.create(v.getString().length());
        }
        if (t == ARRAY) {
            return IntValue.create(v.getList().size());
        }
        throw new InvalidValueTypeError(t, STRING, ARRAY);
    }

    private static Value invokeArray2(VList args) throws VMError {
        args.checkMinCount(2);
        args.checkMaxCount(3);
        int size1 = args.get(0).getInt();
        int size2 = args.get(0).getInt();
        if (size1 < 0 || size2 < 0) {
            throw new InvalidArgumentError("Size must be positive integer");
        }
        if (size1 * size2 > MAX_ARRAY_SIZE) {
            throw new InvalidArgumentError("Too big array size");
        }
        Value v = args.size() >= 3 ? args.get(2) : NullValue.NULL;
        List<Value> list = new ArrayList<Value>(size1);
        for (int i = 0; i < size1; ++i) {
            list.add(createArray(size2, v));
        }
        return ArrayValue.create(list);
    }

    private static Value invokeToInt(VList args) throws VMError {
        args.checkCount(1);
        Value v = args.get(0);
        long res = 0;
        switch (v.getType()) {
            case INT:
                return v;
            case FLOAT:
                res = (int) v.getDouble();
                break;
            case BOOL:
                res = v.getBoolean() ? 1 : 0;
                break;
            case STRING:
                try {
                    res = Integer.parseInt(v.getString());
                } catch (NumberFormatException e) {
                    throw new InvalidArgumentError(
                            "Invalid number \"" + v.getString() + '"');
                }
                break;
            default:
                throw new InvalidValueTypeError(
                        v.getType(), INT, FLOAT, BOOL, STRING);
        }
        return IntValue.create(res);
    }

    private static Value invokeToFloat(VList args) throws VMError {
        args.checkCount(1);
        Value v = args.get(0);
        double res = 0;
        switch (v.getType()) {
            case INT:
                res = v.getLong();
                break;
            case FLOAT:
                return v;
            case STRING:
                try {
                    res = Double.parseDouble(v.getString());
                } catch (NumberFormatException e) {
                    throw new InvalidArgumentError(
                            "Invalid number \"" + v.getString() + '"');
                }
                break;
            default:
                throw new InvalidValueTypeError(
                        v.getType(), INT, FLOAT, STRING);
        }
        return FloatValue.create(res);
    }

    private static Value invokeResize(VList args) throws VMError {
        args.checkMinCount(2);
        args.checkMaxCount(3);
        List<Value> list = args.get(0).getList();
        int newSize = args.get(1).getInt();
        if (newSize < 0) {
            throw new InvalidArgumentError("Size must be positive integer");
        }
        if (newSize > MAX_ARRAY_SIZE) {
            throw new InvalidArgumentError("Too big array size");
        }
        Value v = args.size() >= 3 ? args.get(2) : NullValue.NULL;
        while (list.size() < newSize) {
            list.add(v);
        }
        while (list.size() > newSize) {
            list.remove(list.size() - 1);
        }
        return null;
    }

    private static Value invokeType(VList args) throws VMError {
        args.checkCount(1);
        Value v = args.get(0);
        return StringValue.create(v.getType().toString().toLowerCase());
    }

    private static Value invokeAbs(VList args) throws VMError {
        Value v = args.get(0);
        if (v.getType() == ValueType.INT) {
            return IntValue.create(Math.abs(v.getLong()));
        } else if (v.getType() == ValueType.FLOAT) {
            return FloatValue.create(Math.abs(v.getDouble()));
        } else {
            throw new InvalidValueTypeError(v.getType(), INT, FLOAT);
        }
    }
}
