/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Zhlobich Andrei <mailto:anjensan@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.lang.modules;

import by.informatics.jcx.fingy.lang.*;

public class MathFunction implements Function
{
    private MathFunctionType type;

    public  MathFunction(MathFunctionType type) {
        this.type = type;
    }

    @Override
    public Value invoke(VMContext ctx, VList args) throws VMError {
        if (type.argcount != MathFunctionType.ANY_ARGS_COUNT) {
            args.checkCount(type.argcount);
        }

        double a = Double.NaN;
        double b = Double.NaN;

        if (type.real) {
            if (args.size() >= 1) {
                a = args.get(0).getDouble();
            }
            if (args.size() >= 2) {
                b = args.get(1).getDouble();
            }
        }

        switch (type) {
            case ACOS:
                return FloatValue.create(Math.acos(a));
            case ASIN:
                return FloatValue.create(Math.asin(a));
            case ATAN:
                return FloatValue.create(Math.atan(a));
            case CEIL:
                return FloatValue.create(Math.ceil(a));
            case COS:
                return FloatValue.create(Math.cos(a));
            case EXP:
                return FloatValue.create(Math.exp(a));
            case FLOOR:
                return FloatValue.create(Math.floor(a));
            case LN:
                return FloatValue.create(Math.log(a));
            case LOG:
                return FloatValue.create(Math.log10(a));
            case POW:
                return FloatValue.create(Math.pow(a, b));
            case SIN:
                return FloatValue.create(Math.sin(a));
            case SQRT:
                return FloatValue.create(Math.sqrt(a));
            case TAN:
                return FloatValue.create(Math.tan(a));
            // type.real == false
            case MAX:
                return calcMax(args);
            case MIN:
                return calcMin(args);
        }
        return NullValue.NULL;
    }

    private static Value calcMax(VList args) throws VMError {
        args.checkMinCount(1);
        if (allIsTypeOf(args, ValueType.INT)) {
            long r = args.get(0).getInt();
            for (int i = 1; i < args.size(); ++i) {
                long e = args.get(i).getLong();
                if (e > r) {
                    r = e;
                }
            }
            return IntValue.create(r);
        }
        double r = args.get(0).getDouble();
        for (int i = 1; i < args.size(); ++i) {
            double e = args.get(i).getDouble();
            if (e > r) {
                r = e;
            }
        }
        return FloatValue.create(r);
    }

    private static Value calcMin(VList args) throws VMError {
        args.checkMinCount(1);
        if (allIsTypeOf(args, ValueType.INT)) {
            long r = args.get(0).getInt();
            for (int i = 1; i < args.size(); ++i) {
                long e = args.get(i).getLong();
                if (e < r) {
                    r = e;
                }
            }
            return IntValue.create(r);
        }
        double r = args.get(0).getDouble();
        for (int i = 1; i < args.size(); ++i) {
            double e = args.get(i).getDouble();
            if (e < r) {
                r = e;
            }
        }
        return FloatValue.create(r);
    }

    private static boolean allIsTypeOf(VList args, ValueType type)
            throws VMError {
        for (int i = 0; i < args.size(); ++i) {
            if (args.get(i).getType() != type) {
                return false;
            }
        }
        return true;
    }
}