/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Zhlobich Andrei <mailto:anjensan@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.lang.modules;

import by.informatics.jcx.fingy.lang.ArrayValue;
import by.informatics.jcx.fingy.lang.BoolValue;
import by.informatics.jcx.fingy.lang.Function;
import by.informatics.jcx.fingy.lang.InvalidArgumentError;
import by.informatics.jcx.fingy.lang.Module;
import by.informatics.jcx.fingy.lang.StringValue;
import by.informatics.jcx.fingy.lang.VList;
import by.informatics.jcx.fingy.lang.VMContext;
import by.informatics.jcx.fingy.lang.VMError;
import by.informatics.jcx.fingy.lang.Value;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

public class StringModule implements Module {

    private static Map<String, Function> functions =
            new HashMap<String, Function>();

    static {
        add("substr", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeSubstr(args);
            }
        });
        add("str_to_array", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeStrToArray(args);
            }
        });
        add("join", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                return invokeJoinStr(args);
            }
        });
        add("upper", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                args.checkCount(1);
                return StringValue.create(
                        args.get(0).getString().toUpperCase());
            }
        });
        add("lower", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                args.checkCount(1);
                return StringValue.create(
                        args.get(0).getString().toLowerCase());
            }
        });
        add("trim", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                args.checkCount(1);
                return StringValue.create(
                        args.get(0).getString().trim());
            }
        });
        add("startswith", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                args.checkCount(2);
                String s1 = args.get(0).toString();
                String s2 = args.get(1).toString();
                return BoolValue.create(s1.startsWith(s2));
            }
        });
        add("endswith", new Function() {
            @Override
            public Value invoke(VMContext ctx, VList args) throws VMError {
                args.checkCount(2);
                String s1 = args.get(0).toString();
                String s2 = args.get(1).toString();
                return BoolValue.create(s1.endsWith(s2));
            }
        });
    }

    @Override
    public Function getFunction(String name) {
        return functions.get(name);
    }

    @Override
    public Value getValue(String name) {
        return null;
    }

    @Override
    public String[] getFunctionNames() {
        return functions.keySet().toArray(new String[0]);
    }

    @Override
    public String[] getValueNames() {
        return null;
    }

    public static String[] getNames() {
        return functions.keySet().toArray(new String[0]);
    }

    private static void add(String name, Function function) {
        functions.put(name, function);
    }

    private static Value invokeSubstr(VList args) throws VMError {
        args.checkMinCount(2);
        args.checkMaxCount(3);
        String v = args.get(0).getString();
        try {
            if (args.size() == 2) {
                v = v.substring(args.get(1).getInt());
            } else {
                v = v.substring(args.get(1).getInt(), args.get(2).getInt());
            }
        } catch (IndexOutOfBoundsException e) {
            throw new InvalidArgumentError(e.getMessage());
        }
        return StringValue.create(v);
    }

    private static Value invokeJoinStr(VList args) throws VMError {
        args.checkCount(1);
        StringBuilder sb = new StringBuilder();
        for (Value v : args.get(0).getList()) {
            String s = v.getString();
            sb.append(s);
        }
        return StringValue.create(sb.toString());
    }

    private static Value invokeStrToArray(VList args) throws VMError {
        args.checkCount(1);
        String s = args.get(0).getString();
        ArrayList<Value> list = new ArrayList<Value>();
        for (int i = 0; i < s.length(); ++i) {
            list.add(StringValue.create(String.valueOf(s.charAt(i))));
        }
        return ArrayValue.create(list);
    }
}
