/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Zhlobich Andrei <mailto:anjensan@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.lang.vm;

import by.informatics.jcx.fingy.lang.ConstAssignmentError;
import by.informatics.jcx.fingy.lang.Function;
import by.informatics.jcx.fingy.lang.IdentifierRedeclarationError;
import by.informatics.jcx.fingy.lang.Module;
import by.informatics.jcx.fingy.lang.StringValue;
import by.informatics.jcx.fingy.lang.SyntaxError;
import by.informatics.jcx.fingy.lang.UndeclaredIdentifierError;
import by.informatics.jcx.fingy.lang.UnknownModuleError;
import by.informatics.jcx.fingy.lang.Value;
import by.informatics.jcx.fingy.lang.ValueType;
import by.informatics.jcx.fingy.lang.parser.*;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.EnumMap;
import java.util.List;

class CodeBuilderVisitor implements Visitor {
    
    private static class FunBodyNode extends ASTNode {
        
        private ASTFunDefSt parent;

        private FunBodyNode(ASTFunDefSt funDef) {
            super(funDef);
            this.parent = funDef;
        }

        @Override
        public void visit(Visitor visitor) {
            CodeBuilderVisitor v = (CodeBuilderVisitor) visitor;
            v.buildFunctionCode(parent);
            parent.body.visit(visitor);
        }
    }
    
    private static final Object CONST_TAG = new Object();
    
    private CodeGenerator cg;
    private ConstPoolBuilder cp;
    private NamespaceBuilder nm;
    private int lastLine = -1;
    private Deque<ArgCodeProvider> startLabels =
            new ArrayDeque<ArgCodeProvider>();
    private Deque<ArgCodeProvider> endLabels =
            new ArrayDeque<ArgCodeProvider>();
    private List<SyntaxError> errors = new ArrayList<SyntaxError>();
    private EnumMap<ExprType, Operation> exprOps =
            new EnumMap<ExprType, Operation>(ExprType.class);
    private EnumMap<ExprType, Operation> notExprOps =
            new EnumMap<ExprType, Operation>(ExprType.class);

    {
        for (ExprType expr : ExprType.values()) {
            String ename = expr.toString();
            try {
                Operation op = Operation.valueOf(ename);
                exprOps.put(expr, op);
            } catch (IllegalArgumentException e) {
                // do nothing
            }
        }

        notExprOps.put(ExprType.NE, Operation.EQ);
        notExprOps.put(ExprType.GE, Operation.LT);
        notExprOps.put(ExprType.LE, Operation.GT);
    }

    public CodeBuilderVisitor(
            CodeGenerator codeGenerator,
            ConstPoolBuilder constPoolBuilder,
            NamespaceBuilder namespaceBuilder) {
        this.cg = codeGenerator;
        this.cp = constPoolBuilder;
        this.nm = namespaceBuilder;
    }

    public void visit(ASTNode node) {
        throw new AbstractMethodError("Visit unknown AST-node.");
    }

    public void visit(ASTOpExpr node) {
        visitAll(node.args);
        ExprType t = node.type;
        if (exprOps.containsKey(t)) {
            cg.addOp(exprOps.get(t));
        } else if (notExprOps.containsKey(t)) {
            cg.addOp(notExprOps.get(t));
            cg.addOp(Operation.NOT);
        } else {
            throw new AbstractMethodError("Unknown ASTOpExpr node.");
        }
    }

    public void visit(ASTConstExpr node) {
        ArgCodeProvider cid = cp.add(node.value);
        cg.addOp(Operation.CLOAD, cid);
    }

    public void visit(ASTVariableExpr node) {
        if (!nm.containsName(node.variable)) {
            errors.add(new UndeclaredIdentifierError(node.line, node.variable));
            return;
        } else {
            ArgCodeProvider cid = nm.getNameCode(node.variable);
            cg.addOp(Operation.VLOAD, cid);
        }
    }

    public void visit(ASTFunCallExpr node) {
        for (ASTExpr e : node.args) {
            e.visit(this);
        }
        node.funid.visit(this);
        cg.addOp(Operation.FARG, node.args.size());
        cg.addOp(Operation.FCALL);
    }

    public void visit(ASTIndexedExpr node) {
        node.array.visit(this);
        visitAll(node.indexes);
        cg.addOp(Operation.FARG, node.indexes.size());
        cg.addOp(Operation.ICALL, ArrayHelper.GET_CALL_ID);
    }

    public void visit(ASTNewArrayExpr node) {
        processLine(node);
        cg.addOp(Operation.ICALL, ArrayHelper.INIT_CALL_ID);
        for (ASTExpr e : node.values) {
            e.visit(this);
            cg.addOp(Operation.ICALL, ArrayHelper.APPEND_CALL_ID);
        }
    }

    public void visit(ASTStatement node) {
        throw new AbstractMethodError("Visit unknown AST-node.");
    }

    public void visit(ASTConditionalSt node) {
        processLine(node);
        ArgCodeProvider lb = cg.newLabel();
        ArgCodeProvider end = cg.newLabel();
        for (int i = 0; i < node.length; ++i) {
            ASTExpr cond = node.conditions.get(i);
            ASTStatement st = node.statements.get(i);
            processLine(node);
            cond.visit(this);
            cg.addOp(Operation.GONIF, lb);
            st.visit(this);
            cg.addOp(Operation.GOTO, end);
            cg.setLabel(lb);
            lb = cg.newLabel();
        }
        if (node.isElse) {
            node.statements.get(node.length).visit(this);
        }
        cg.setLabel(end);
    }

    public void visit(ASTForLoopSt node) {
        processLine(node);
        if (nm.containsLocalName(node.variable)) {
            errors.add(new IdentifierRedeclarationError(
                    node.line, node.variable));
            return;
        }
        pushNames();
        addName(node, node.variable);
        addNameTag(node.variable);
        ArgCodeProvider varid = nm.getNameCode(node.variable);
        ArgCodeProvider start = cg.newLabel();
        ArgCodeProvider end = cg.newLabel();
        startLabels.push(start);
        endLabels.push(end);

        node.from.visit(this);
        cg.addOp(Operation.DUP);
        cg.addOp(Operation.ICALL, ForLoopHelper.FROM_CALL_ID);
        node.to.visit(this);
        cg.addOp(Operation.ICALL, ForLoopHelper.TO_CALL_ID);
        node.by.visit(this);
        cg.addOp(Operation.ICALL, ForLoopHelper.BY_CALL_ID);
        cg.addOp(Operation.DUP);
        cg.addOp(Operation.VSTOR, varid);
        cg.addOp(Operation.ICALL, ForLoopHelper.CHECK_CALL_ID);
        cg.addOp(Operation.GONIF, end);
        cg.setLabel(start);
        node.body.visit(this);

        processLine(node, true);  // add LINE op
        cg.addOp(Operation.VLOAD, varid);
        cg.addOp(Operation.ICALL, ForLoopHelper.INC_CALL_ID);
        cg.addOp(Operation.DUP);
        cg.addOp(Operation.VSTOR, varid);
        cg.addOp(Operation.ICALL, ForLoopHelper.CHECK_CALL_ID);
        cg.addOp(Operation.GOIF, start);

        cg.setLabel(end);
        cg.addOp(Operation.ICALL, ForLoopHelper.ENDFOR_CALL_ID);

        popNames();

        startLabels.pop();
        endLabels.pop();
    }

    public void visit(ASTWhileLoopSt node) {
        ArgCodeProvider start, end;
        start = cg.newLabel();
        end = cg.newLabel();
        startLabels.push(start);
        endLabels.push(end);

        cg.setLabel(start);
        processLine(node);
        node.condition.visit(this);
        cg.addOp(Operation.GONIF, end);
        node.body.visit(this);
        cg.addOp(Operation.GOTO, start);

        cg.setLabel(end);
        startLabels.pop();
        endLabels.pop();
    }

    public void visit(ASTInfiniteLoopSt node) {
        processLine(node);
        ArgCodeProvider start = cg.newLabel();
        ArgCodeProvider end = cg.newLabel();
        startLabels.push(start);
        endLabels.push(end);

        cg.setLabel(start);
        node.body.visit(this);
        cg.addOp(Operation.GOTO, start);

        cg.setLabel(end);
        startLabels.pop();
        endLabels.pop();
    }

    public void visit(ASTVarDeclSt node) {
        processLine(node);
        if (nm.containsLocalName(node.name)) {
            errors.add(new IdentifierRedeclarationError(node.line, node.name));
            return;
        }
        if (node.value != null) {
            node.value.visit(this);
        } else {
            cg.addOp(Operation.CLOAD, Const.NULL);
        }
        addName(node, node.name);
        ArgCodeProvider vid = nm.getNameCode(node.name);
        addNameTag(node.name);
        cg.addOp(Operation.VSTOR, vid);
    }

    public void visit(ASTAssignmentSt node) {
        processLine(node);
        if (!nm.containsName(node.variable)) {
            errors.add(new UndeclaredIdentifierError(node.line, node.variable));
            return;
        }
        if (isConstTag(nm.getTag(node.variable))) {
            errors.add(new ConstAssignmentError(node.line, node.variable));
            return;
        }
        ArgCodeProvider vid = nm.getNameCode(node.variable);
        node.value.visit(this);
        if (node.indexes == null) {
            cg.addOp(Operation.VSTOR, vid);
        } else {
            cg.addOp(Operation.VLOAD, vid);
            visitAll(node.indexes);
            cg.addOp(Operation.FARG, node.indexes.size());
            cg.addOp(Operation.ICALL, ArrayHelper.SET_CALL_ID);
            cg.addOp(Operation.NOTIFY, vid);
        }
    }

    public void visit(ASTBlockSt node) {
        pushNames();
        visitAll(node.body);
        popNames();
    }

    public void rethrowErrors() throws SyntaxError {
        if (!errors.isEmpty()) {
            throw errors.get(0);
        }
    }

    private final <Node extends ASTNode> void visitAll(List<Node> nodes) {
        for (Node n : nodes) {
            n.visit(this);
        }
    }

    private final void processLine(ASTNode node) {
        processLine(node, false);
    }

    private final void processLine(ASTNode node, boolean force) {
        if (node.line != -1 && (lastLine < node.line || force)) {
            lastLine = node.line;
            cg.addOp(Operation.LINE, lastLine);
        }
    }

    private final void addName(ASTNode node, String name) {
        if (nm.containsLocalName(name)) {
            errors.add(new IdentifierRedeclarationError(node.line, name));
        } else {
            nm.addName(name);
        }
    }

    public void visit(ASTFunCallSt node) {
        processLine(node);
        node.funid.visit(this);
        cg.addOp(Operation.POP);
    }

    public void visit(ASTFunDefSt node) {
        processLine(node);
        boolean containsName = nm.containsLocalName(node.name);
        boolean defined = containsName && nm.getTag(node.name) != null;
        boolean predefined = containsName && nm.getTag(node.name) == null;

        if (defined || predefined && node.ispre) {
            errors.add(new IdentifierRedeclarationError(node.line, node.name));
            return;
        }
        if (!containsName) {
            addName(node, node.name);
        }

        ArgCodeProvider vFunid = nm.getNameCode(node.name);
        NamespaceBuilder subNM = new NamespaceBuilder(nm);

        if (node.ispre) {
            return;
        }

        VMCode code;
        try {
            code = VMCode.create(new FunBodyNode(node), subNM, cp);
        } catch (SyntaxError error) {
            errors.add(error);
            return;
        }
        processLine(node);
        UserFunctionPrototype ufs =
                new UserFunctionPrototype(node.args, code);
        InternalValue uf =
                new InternalValue(ufs, InternalValueType.UFUN_PROTOTYPE);
        ArgCodeProvider funSkelId = cp.add(uf);
        cg.addOp(Operation.FARG, funSkelId);
        cg.addOp(Operation.FDEF, vFunid);
        nm.addTag(node.name, ufs);
    }

    private void buildFunctionCode(ASTFunDefSt node) {
        processLine(node);
        int n = node.args.size();
        for (int i = 0; i < n; ++i) {
            final String name = node.args.get(i);
            if (nm.containsLocalName(name)) {
                errors.add(new IdentifierRedeclarationError(node.line, node.name));
                return;
            }
            addName(node, name);
        }
        for (int i = n - 1; i >= 0; --i) {
            final String name = node.args.get(i);
            ArgCodeProvider vid = nm.getNameCode(name);
            addNameTag(name);
            cg.addOp(Operation.VSTOR, vid);
        }
    }

    public void visit(ASTReturnSt node) {
        processLine(node);
        if (node.value == null) {
            cg.addOp(Operation.CLOAD, Const.NULL);
        } else {
            node.value.visit(this);
        }
        cg.addOp(Operation.RET);
    }

    public void visit(ASTLoopControlSt node) {
        ArgCodeProvider label;
        switch (node.kind) {
            case ASTLoopControlSt.BREAK:
                label = endLabels.peek();
                break;
            case ASTLoopControlSt.CONTINUE:
                label = startLabels.peek();
                break;
            default:
                throw new IllegalArgumentException(
                        "Unknown loop-control-statement code");
        }
        if (label == null) {
            errors.add(new SyntaxError(
                    node.line, "Control statement outside loop"));
        } else {
            cg.addOp(Operation.GOTO, label);
        }
    }

    private void addNameTag(String name) {
        cg.addOp(Operation.NAME, cp.add(StringValue.create(name)));
    }

    private void removeNameTag(String name) {
        ArgCodeProvider id = nm.getNameCode(name);
        cg.addOp(Operation.UVAR, id);
    }

    private void pushNames() {
        nm.pushState();
    }

    private void popNames() {
        for (String name : nm.peekState()) {
            removeNameTag(name);
        }
        nm.popState();
    }

    public void visit(ASTModuleImportSt node) {
        processLine(node);

        if (!nm.containsName("module$" + node.module)) {
            errors.add(new UnknownModuleError(this.lastLine, node.module));
            return;
        }

        if (!nm.containsName("module_imported$" + node.module)) {
            nm.addName("module_imported$" + node.module);
            // module already imported...
        }

        String mname = "module$" + node.module;

        Module m;
        {
            Value mv = nm.getNameValue(mname);
            if (mv.getType() != ValueType.INTERNAL) {
                throw new InternalError("Invalid internal value");
            }
            InternalValue iv = (InternalValue) mv;
            if (iv.type != InternalValueType.MODULE) {
                throw new InternalError("Invalid internal value");
            }
            m = (Module) iv.getObject();
        }

        String[] names;
        names = m.getFunctionNames();
        if (names != null) {
            for (String name : m.getFunctionNames()) {
                Function f = m.getFunction(name);
                ExternalFunction ef = new ExternalFunction(f);
                Value v = new InternalValue(ef, InternalValueType.EFUNCTION);
                tryToAddName(name, v);
            }
        }

        names = m.getValueNames();
        if (names != null) {
            for (String name : names) {
                Value v = m.getValue(name);
                tryToAddName(name, v);
            }
        }
    }

    private void tryToAddName(String name, Value value) {
        if (nm.containsLocalName(name)) {
            ArgCodeProvider cid = nm.getNameCode(name);
            cg.addOp(Operation.CLOAD, cid);
        }
        else {
            nm.addName(name, value);
            nm.addTag(name, CONST_TAG);
        }
    }

    @Override
    public void visit(ASTTimesLoopSt node) {
        processLine(node);
        ArgCodeProvider start = cg.newLabel();
        ArgCodeProvider end = cg.newLabel();
        startLabels.push(start);
        endLabels.push(end);

        node.times.visit(this);
        cg.addOp(Operation.ICALL, ForLoopHelper.C_INT_CALL_ID);
        cg.addOp(Operation.CLOAD, Const.ONE);
        cg.addOp(Operation.ICALL, ForLoopHelper.FROM_CALL_ID);
        cg.addOp(Operation.ICALL, ForLoopHelper.TO_CALL_ID);
        cg.addOp(Operation.CLOAD, Const.ONE);
        cg.addOp(Operation.DUP);
        cg.addOp(Operation.ICALL, ForLoopHelper.BY_CALL_ID);
        cg.addOp(Operation.DUP);
        cg.addOp(Operation.ICALL, ForLoopHelper.CHECK_CALL_ID);
        cg.addOp(Operation.GONIF, end);
        cg.setLabel(start);
        node.body.visit(this);
        processLine(node, true);  // add LINE op
        cg.addOp(Operation.ICALL, ForLoopHelper.INC_CALL_ID);
        cg.addOp(Operation.DUP);
        cg.addOp(Operation.ICALL, ForLoopHelper.CHECK_CALL_ID);
        cg.addOp(Operation.GOIF, start);
        cg.setLabel(end);
        cg.addOp(Operation.ICALL, ForLoopHelper.ENDFOR_CALL_ID);
        cg.addOp(Operation.POP);

        startLabels.pop();
        endLabels.pop();
    }

    @Override
    public void visit(ASTForInLoopSt node) {
        processLine(node);

        if (nm.containsLocalName(node.variable)) {
            errors.add(new IdentifierRedeclarationError(
                    node.line, node.variable));
            return;
        }
        pushNames();
        addName(node, node.variable);
        addNameTag(node.variable);
        ArgCodeProvider varid = nm.getNameCode(node.variable);
        ArgCodeProvider start = cg.newLabel();
        ArgCodeProvider end = cg.newLabel();
        startLabels.push(start);
        endLabels.push(end);

        node.sequence.visit(this);
        cg.addOp(Operation.ICALL, ForLoopHelper.C_ARRAY_CALL_ID);
        cg.addOp(Operation.CLOAD, Const.ZERO);
        cg.addOp(Operation.ICALL, ForLoopHelper.FROM_CALL_ID);
        cg.addOp(Operation.CLOAD, Const.ONE);
        cg.addOp(Operation.ICALL, ForLoopHelper.BY_CALL_ID);
        cg.addOp(Operation.CLOAD, Const.ZERO);
        cg.addOp(Operation.ICALL, ForLoopHelper.CHECKA_CALL_ID);
        cg.addOp(Operation.GONIF, end);
        cg.setLabel(start);
        cg.addOp(Operation.ICALL, ForLoopHelper.DDUP_CALL_ID);
        cg.addOp(Operation.FARG, 1);
        cg.addOp(Operation.ICALL, ArrayHelper.GET_CALL_ID);
        cg.addOp(Operation.VSTOR, varid);
        node.body.visit(this);
        processLine(node, true);  // add LINE op
        cg.addOp(Operation.ICALL, ForLoopHelper.INC1_CALL_ID);
        cg.addOp(Operation.ICALL, ForLoopHelper.CHECKA_CALL_ID);
        cg.addOp(Operation.GOIF, start);
        cg.setLabel(end);
        cg.addOp(Operation.ICALL, ForLoopHelper.ENDFORA_CALL_ID);
        cg.addOp(Operation.POP);
        cg.addOp(Operation.POP);

        popNames();

        startLabels.pop();
        endLabels.pop();
    }

    @Override
    public void visit(ASTConstDeclSt node) {
        processLine(node);
        if (nm.containsLocalName(node.name)) {
            errors.add(new IdentifierRedeclarationError(node.line, node.name));
            return;
        }
        addNameTag(node.name);
        nm.addTag(node.name, CONST_TAG);
        node.value.visit(this);
        addName(node, node.name);
        ArgCodeProvider vid = nm.getNameCode(node.name);
        cg.addOp(Operation.VSTOR, vid);
    }

    private static boolean isConstTag(Object tag) {
        return tag == CONST_TAG || tag instanceof UserFunctionPrototype;
    }
}
