/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Zhlobich Andrei <mailto:anjensan@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.lang.vm;

import by.informatics.jcx.fingy.lang.IntValue;
import by.informatics.jcx.fingy.lang.InvalidValueTypeError;
import by.informatics.jcx.fingy.lang.VMError;
import by.informatics.jcx.fingy.lang.Value;
import by.informatics.jcx.fingy.lang.ValueType;
import java.util.ArrayDeque;
import java.util.Deque;

class ForLoopHelper implements InternalFunction {

    public static final int FROM_CALL_ID = 10;
    public static final int TO_CALL_ID = 11;
    public static final int BY_CALL_ID = 12;
    public static final int INC_CALL_ID = 13;
    public static final int CHECK_CALL_ID = 14;
    public static final int ENDFOR_CALL_ID = 15;
    public static final int C_INT_CALL_ID = 16;
    public static final int C_ARRAY_CALL_ID = 17;
    public static final int CHECKA_CALL_ID = 18;
    public static final int DDUP_CALL_ID = 19;
    public static final int ENDFORA_CALL_ID = 9;
    public static final int INC1_CALL_ID = 8;

    private Deque<Value> bySt = new ArrayDeque<Value>();
    private Deque<Value> toSt = new ArrayDeque<Value>();

    private ForLoopHelper(VMExecutor e) {
        e.addIFunc(this, FROM_CALL_ID);
        e.addIFunc(this, TO_CALL_ID);
        e.addIFunc(this, BY_CALL_ID);
        e.addIFunc(this, INC_CALL_ID);
        e.addIFunc(this, CHECK_CALL_ID);
        e.addIFunc(this, CHECKA_CALL_ID);
        e.addIFunc(this, ENDFOR_CALL_ID);
        e.addIFunc(this, C_INT_CALL_ID);
        e.addIFunc(this, C_ARRAY_CALL_ID);
        e.addIFunc(this, DDUP_CALL_ID);
        e.addIFunc(this, ENDFORA_CALL_ID);
        e.addIFunc(this, INC1_CALL_ID);
    }

    public void invoke(int code, VMExecutor e) throws VMError {
        switch (code) {
            case BY_CALL_ID:
                bySt.add(e.pop());
                break;
            case FROM_CALL_ID:
                e.pop();
                break;
            case TO_CALL_ID:
                toSt.add(e.pop());
                break;
            case INC1_CALL_ID:
                e.push(IntValue.create(e.pop().getInt() + 1));
                break;
            case INC_CALL_ID:
                e.push(bySt.peek());
                e.command(Operation.ADD);
                break;
            case CHECK_CALL_ID:
                e.push(toSt.peekLast());
                if (bySt.peekLast().getDouble() > 0) {
                    e.command(Operation.GT);
                } else {
                    e.command(Operation.LT);
                }
                e.command(Operation.NOT);
                break;
            case CHECKA_CALL_ID:
                e.command(Operation.SWAP);
                int len = e.peek().getList().size();
                e.command(Operation.SWAP);
                e.command(Operation.DUP);
                e.push(IntValue.create(len));
                e.command(Operation.LT);
                break;
            case ENDFOR_CALL_ID:
                toSt.removeLast();
                bySt.removeLast();
                break;
            case ENDFORA_CALL_ID:
                break;
            case C_INT_CALL_ID:
                if (e.peek().getType() != ValueType.INT) {
                    throw new InvalidValueTypeError(
                            e.peek().getType(), ValueType.INT);
                }
                break;
            case C_ARRAY_CALL_ID:
                if (e.peek().getType() != ValueType.ARRAY) {
                    throw new InvalidValueTypeError(
                            e.peek().getType(), ValueType.ARRAY);
                }
                break;
            case DDUP_CALL_ID: {
                Value v1 = e.pop();
                Value v2 = e.peek();
                e.push(v1);
                e.push(v2);
                e.push(v1);
            }
            break;

            default:
                throw new InternalVMError("Invalid ForLoopHelper code");
        }
    }

    public static void register(VMExecutor e) {
        new ForLoopHelper(e);
    }
}
