/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Zhlobich Andrei <mailto:anjensan@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.lang.vm;

import by.informatics.jcx.fingy.lang.Value;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Deque;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

class NamespaceBuilder {

    public static final int SHIFT_DELTA = 100000;

    public final NamespaceBuilder parent;
    private Map<String, Integer> npool = new HashMap<String, Integer>();
    private Map<String, Object> tags = new HashMap<String, Object>();
    private Map<String, Value> addedValues = new HashMap<String, Value>();
    private Deque<List<String>> nStack = new ArrayDeque<List<String>>();
    private int curId = 0;
    private int maxId = 0;
    private List<Value> initvals = new ArrayList<Value>();
    private List<Integer> initids = new ArrayList<Integer>();
    private int shift;

    private static class NameArgCodeProvider implements ArgCodeProvider {

        private final int code;
        private final String name;

        public NameArgCodeProvider(int code, String name) {
            this.code = code;
            this.name = name;
        }

        public int getCode() {
            return code;
        }

        @Override
        public String toString() {
            return String.format("V%1d  %1s", code, name);
        }
    }

    public NamespaceBuilder(NamespaceBuilder parent) {
        this.pushState();
        this.parent = parent;
        if (parent != null) {
            shift = parent.shift + SHIFT_DELTA;
        }
    }

    public NamespaceBuilder() {
        this(null);
        shift = 0;
    }

    public void addName(String name) {
        if (containsLocalName(name)) {
            throw new java.lang.IllegalArgumentException();
        }
        npool.put(name, curId++);
        if (curId > maxId) {
            maxId = curId;
        }
        nStack.peek().add(name);
    }

    public void addTag(String name, Object tag) {
        tags.put(name, tag);
    }

    public Object getTag(String name) {
        if (containsLocalName(name)) {
            return tags.get(name);
        } else if (parent != null) {
            return parent.getTag(name);
        }
        return null;
    }

    public void addName(String name, Value value) {
        addName(name);
        initids.add(npool.get(name));
        initvals.add(value);
        addedValues.put(name, value);
    }

    public Value getNameValue(String name) {
        if (!containsName(name)) {
            return null;
        }
        if (containsLocalName(name)) {
            return addedValues.get(name);
        } else {
            return this.parent.getNameValue(name);
        }
    }

    public void pushState() {
        nStack.push(new ArrayList<String>());
    }

    public List<String> peekState() {
        return nStack.peek();
    }

    public List<String> popState() {
        List<String> res = nStack.pop();
        for (String name : res) {
            npool.remove(name);
            tags.remove(name);
        }
        return res;
    }

    public boolean containsLocalName(String name) {
        return npool.containsKey(name);
    }

    public boolean containsName(String name) {
        return containsLocalName(name)
                || parent != null && parent.containsName(name);
    }

    public ArgCodeProvider getNameCode(String name) {
        if (containsLocalName(name)) {
            int code = npool.get(name) + shift;
            return new NameArgCodeProvider(code, name);
        } else {
            return parent.getNameCode(name);
        }
    }

    public VarPoolBuilder createPoolBuilder() {
        if (initvals.isEmpty()) {
            return new VarPoolBuilder(maxId, shift);
        } else {
            return new VarPoolBuilder(maxId, shift, initids, initvals);
        }
    }
}
