/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Zhlobich Andrei <mailto:anjensan@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.lang.vm;

import by.informatics.jcx.fingy.lang.DivisionByZeroError;
import by.informatics.jcx.fingy.lang.FloatValue;
import by.informatics.jcx.fingy.lang.IntValue;
import by.informatics.jcx.fingy.lang.InvalidParametersCountError;
import by.informatics.jcx.fingy.lang.VMStackOverflowError;
import by.informatics.jcx.fingy.lang.VMError;
import by.informatics.jcx.fingy.lang.Value;
import by.informatics.jcx.fingy.lang.ValueType;
import by.informatics.jcx.fingy.lang.InvalidValueTypeError;
import by.informatics.jcx.fingy.lang.NullValue;
import by.informatics.jcx.fingy.lang.StringValue;
import by.informatics.jcx.fingy.lang.VList;
import by.informatics.jcx.fingy.lang.Variable;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Deque;
import java.util.List;

final class VMExecutor {

    // some constants
    private static final int MAXIFUNCTIONS = 32;
    // registers
    private int[] code;
    private VarPool varp;
    private int id;
    private int rLine;
    private int rFarg;

    private final class RegistersSnapshot {

        private int[] code;
        private VarPool varp;
        private int id;
        private int rLine;
        private int rFarg;

        public RegistersSnapshot() {
            VMExecutor e = VMExecutor.this;
            code = e.code;
            varp = e.varp;
            id = e.id;
            rLine = e.rLine;
            rFarg = e.rFarg;
        }

        public void restore() {
            VMExecutor e = VMExecutor.this;
            e.code = code;
            e.varp = varp;
            e.id = id;
            e.rLine = rLine;
            e.rFarg = rFarg;
        }
    }

    // persistent data
    private InternalFunction[] ifunctions = new InternalFunction[MAXIFUNCTIONS];
    private VMMachine machine;
    private static final boolean DEBUG = false;
    private boolean oneLine;
    private static final Operation[] ops = Operation.values();
    private ConstPool constp;
    private Deque<Value> stack = new ArrayDeque<Value>();
    private Deque<RegistersSnapshot> regStack = new ArrayDeque<RegistersSnapshot>();
    private boolean rExit;
    private volatile boolean rPause;
    private boolean rInterruptMainLoop;
    private boolean isStopped;
    private int rDeep;
    private int rName = -1;
    private final boolean debug;
    private List<Collection<Variable>> variables;
    private final int stackLimit = VMMachine.STACK_LIMIT;

    public VMExecutor(VMMachine machine, VMCode code, boolean debug) {
        this.machine = machine;
        this.code = code.code;
        this.constp = code.cp;
        this.varp = code.vpb.createVarPool();
        this.debug = debug;
        if (debug) {
            this.varp.initVariables();
            this.variables = new ArrayList<Collection<Variable>>();
            pushVariables();
        }
    }

    public void addIFunc(InternalFunction f, int id) {
        ifunctions[id] = f;
    }

    public boolean run() throws VMError {
        oneLine = false;
        runCommands();
        return !rExit;
    }

    public final boolean runLine() throws VMError {
        oneLine = true;
        runCommands();
        return !rExit;
    }

    public final boolean isPaused() {
        return isStopped && rPause;
    }

    private final void runCommands() throws VMError {
        rExit = false;
        rPause = false;
        isStopped = false;
        rInterruptMainLoop = false;
        while (!rInterruptMainLoop) {
            processOneCommand();
        }
        isStopped = true;
        if (rExit && !stack.isEmpty()) {
            printf("Stack:");
            for (Value v : stack) {
                printf("%1s", v);
            }
            throw new InternalVMError("Stack not empty.");
        }
    }

    public int getLine() {
        return rLine;
    }

    public final void command(Operation op) throws VMError {
        command(op, -1);
    }

    public final int getFArg() {
        return this.rFarg;
    }

    public void pause() {
        rPause = true;
    }

    public boolean isStopped() {
        return isStopped;
    }

    int getStackDeep() {
        return rDeep;
    }

    public final void command(Operation op, int arg) throws VMError {
        Value v1, v2;
        switch (op) {
            case GOTO:
                id = arg;
                break;

            case GOIF:
                if (pop().getBoolean()) {
                    id = arg;
                }
                break;

            case GONIF:
                if (!pop().getBoolean()) {
                    id = arg;
                }
                break;

            case RET:
                if (regStack.isEmpty()) {
                    // exit
                    pop(); // skip last result
                    rInterruptMainLoop = true;
                    rExit = true;
                } else {
                    popRegisters();
                    if (debug) {
                        popVariables();
                    }
                }
                rDeep--;
                break;

            case EXIT:
                rInterruptMainLoop = true;
                rExit = true;
                break;

            case CLOAD:
                push(constp.get(arg));
                break;

            case FARG:
                rFarg = arg;
                break;

            case FCALL:
                doFCall(arg);
                break;

            case ICALL:
                ifunctions[arg].invoke(arg, this);
                break;

            case VSTOR:
                if (DEBUG) {
                    printf("V%1d = %2s", arg, peek());
                }
                if (rName != -1) {
                    addName(arg, rName);
                    rName = -1;
                }
                varp.put(arg, pop());
                break;
            
            case NOTIFY:
                if (debug) {
                    varp.put(arg, varp.get(arg));
                }
                break;

            case VLOAD:
                push(varp.get(arg));
                break;

            case LINE:
                rLine = arg;
                if (DEBUG) {
                    printf("%1d:", arg);
                }
                if (rPause || oneLine) {
                    rInterruptMainLoop = true;
                    return;
                }
                break;

            case DUP:
                push(peek());
                break;

            case POP:
                pop();
                break;

            case SWAP:
                v1 = pop();
                v2 = pop();
                push(v1);
                push(v2);
                break;

            // binary operators
            case ADD:
                v1 = pop();
                v2 = pop();
                if (v1.getType() == ValueType.STRING
                        && v2.getType() == ValueType.STRING) {
                    push(opAddString(v1, v2));
                    break;
                }
                push(opArithmetic(op, v1, v2));
                break;
            case SUB:
            case MUL:
            case DIV:
            case POW:
                push(opArithmetic(op, pop(), pop()));
                break;

            // comparators
            case EQ:
                push(opEq(pop(), pop()));
                break;
            case LT:
                push(opLt(pop(), pop()));
                break;
            case GT:
                v1 = pop();
                v2 = pop();
                push(opLt(v2, v1));
                break;

            // unary operators
            case NEG:
                push(opNeg(pop()));
                break;
            case POS:
                push(opPos(pop()));
                break;

            // logical operators
            case NOT:
                push(createBoolValue(!pop().getBoolean()));
                break;
            case AND:
                push(createBoolValue(pop().getBoolean() & pop().getBoolean()));
                break;
            case OR:
                push(createBoolValue(pop().getBoolean() | pop().getBoolean()));
                break;
            case XOR:
                push(createBoolValue(pop().getBoolean() ^ pop().getBoolean()));
                break;

            case FDEF:
                doFDef(arg);
                break;

            case NAME:
                if (debug) {
                    rName = arg;
                }
                break;

            case UVAR:
                if (debug) {
                    removeName(arg);
                }
                break;

            default:
                throw new InternalVMError("Unknown VM opration.");
        }
    }

    private final void pushRegisters() {
        regStack.push(new RegistersSnapshot());
    }

    private final void popRegisters() {
        regStack.pop().restore();
    }

    private final void processOneCommand() throws VMError {
        Operation op = ops[code[id++]];
        int arg;
        if (op.extraArgument) {
            arg = code[id++];
        } else {
            arg = -1;
        }
        command(op, arg);
    }

    public final void push(Value value) throws VMError {
        if (value == null) {
            value = NullValue.NULL;
        }
        stack.push(value);
    }

    public final Value pop() throws VMError {
        return stack.pop();
    }

    public final Value peek() throws VMError {
        return stack.peek();
    }

    private final void printf(String format, Object... args) {
        System.out.printf(format, args);
        System.out.println();
    }

    private static final boolean isNum(ValueType valueType) {
        return valueType == ValueType.INT || valueType == ValueType.FLOAT;
    }

    private final Value createBoolValue(boolean value) {
        return value ? Const.TRUE.constValue : Const.FALSE.constValue;
    }

    // operations
    private final Value opNeg(Value value) throws VMError {
        switch (value.getType()) {
            case INT:
                return IntValue.create(-value.getLong());
            case FLOAT:
                return FloatValue.create(-value.getDouble());
            default:
                throw new InvalidValueTypeError(value.getType(), ValueType.INT,
                        ValueType.FLOAT);
        }
    }

    private final Value opPos(Value value) throws VMError {
        switch (value.getType()) {
            case INT:
            case FLOAT:
                return value;
            default:
                throw new InvalidValueTypeError(value.getType(), ValueType.INT,
                        ValueType.FLOAT);
        }
    }

    private final Value opArithmetic(Operation op, Value v2, Value v1)
            throws VMError {
        ValueType t1 = v1.getType();
        ValueType t2 = v2.getType();
        if (t1 == t2 && t1 == ValueType.INT) {
            return opIntArithmetic(op, v2, v1);
        } else if (isNum(t1) && isNum(t2)) {
            return opFloatArithmetic(op, v2, v1);
        }
        if (isNum(t1)) {
            throw new InvalidValueTypeError(t2, t1);
        } else {
            throw new InvalidValueTypeError(t2, ValueType.STRING);
        }
    }

    private final Value opIntArithmetic(Operation op, Value v2, Value v1)
            throws VMError {
        long a = v1.getLong();
        long b = v2.getLong();
        switch (op) {
            case ADD:
                return IntValue.create(a + b);
            case SUB:
                return IntValue.create(a - b);
            case MUL:
                return IntValue.create(a * b);
            case DIV:
                if (b == 0) {
                    throw new DivisionByZeroError();
                }
                return FloatValue.create((double) a / b);
            case POW:
                return FloatValue.create(Math.pow(a, b));
        }
        throw new InternalVMError("Unknown IntArithmetic operation.");
    }

    private final Value opFloatArithmetic(Operation op, Value v2, Value v1)
            throws VMError {
        double a = v1.getDouble();
        double b = v2.getDouble();
        switch (op) {
            case ADD:
                return FloatValue.create(a + b);
            case SUB:
                return FloatValue.create(a - b);
            case MUL:
                return FloatValue.create(a * b);
            case DIV:
                if (b == 0) {
                    throw new DivisionByZeroError();
                }
                return FloatValue.create(a / b);
            case POW:
                return FloatValue.create(Math.pow(a, b));
        }
        throw new InternalVMError("Unknown FloatArithmetic operation.");
    }

    private final Value opAddString(Value v2, Value v1) throws VMError {
        ValueType t1 = v1.getType();
        ValueType t2 = v2.getType();
        if (t1 == t2 && t1 == ValueType.STRING) {
            return StringValue.create(v1.getString() + v2.getString());
        }
        throw new InvalidValueTypeError(t1, ValueType.STRING);
    }

    private final Value opEq(Value v2, Value v1) throws VMError {
        ValueType t1 = v1.getType();
        ValueType t2 = v2.getType();
        if (t1 == t2 && t1 == ValueType.INT) {
            return createBoolValue(v1.getLong() == v2.getLong());
        } else if (isNum(t1) && isNum(t2)) {
            return createBoolValue(v1.getDouble() == v2.getDouble());
        } else if (t1 == t2 && t1 == ValueType.STRING) {
            return createBoolValue(v1.getString().equals(v1.getString()));
        }

        if (isNum(t1) || t1 == ValueType.STRING) {
            throw new InvalidValueTypeError(t2, t1);
        } else {
            throw new InvalidValueTypeError(t1, ValueType.INT, ValueType.FLOAT,
                    ValueType.STRING);
        }
    }

    private final Value opLt(Value v2, Value v1) throws VMError {
        ValueType t1 = v1.getType();
        ValueType t2 = v2.getType();
        if (t1 == t2 && t1 == ValueType.INT) {
            return createBoolValue(v1.getLong() < v2.getLong());
        } else if (isNum(t1) && isNum(t2)) {
            return createBoolValue(v1.getDouble() < v2.getDouble());
        } else if (t1 == t2 && t1 == ValueType.STRING) {
            return createBoolValue(v1.getString().compareTo(v1.getString()) < 0);
        }

        if (isNum(t1) || t1 == ValueType.STRING) {
            throw new InvalidValueTypeError(t2, t1);
        } else {
            throw new InvalidValueTypeError(t1, ValueType.INT, ValueType.FLOAT,
                    ValueType.STRING);
        }
    }

    private final VList makeVList() throws VMError {
        Value[] v = new Value[rFarg];
        for (int i = 0; i < rFarg; ++i) {
            v[rFarg - i - 1] = pop();
        }
        return new VList(v);
    }

    private final InternalValue toInternalValue(Value v) throws VMError {
        if (v.getType() != ValueType.INTERNAL) {
            throw new InvalidValueTypeError("Callign value as function.");
        }
        return (InternalValue) v;
    }

    private final void doFCall(int arg) throws VMError {
        Value funval = pop();
        InternalValue iv = toInternalValue(funval);
        UserFunction uf;
        Value v;
        switch (iv.type) {
            case EFUNCTION:
                v = ((ExternalFunction) iv.value).invoke(machine, makeVList());
                push(v);
                break;
            case UFUNCTION:
                uf = ((UserFunction) iv.value);
                if (rFarg != uf.argsCount) {
                    throw new InvalidParametersCountError();
                }
                if (rDeep >= stackLimit) {
                    throw new VMStackOverflowError();
                }
                pushRegisters();
                rDeep++;
                code = uf.code.code;
                varp = uf.createVarPool();
                if (debug) {
                    varp.initVariables();
                    pushVariables();                    
                }
                id = 0;
                break;
            default:
                throw new InvalidValueTypeError("Callign value as function.");
        }
    }

    private void pushVariables() {
        variables.add(Collections.unmodifiableCollection(varp.getVariables()));
    }
    
    private void popVariables() {
        variables.remove(variables.size() - 1);
    }

    private final void doFDef(int arg) throws VMError {
        InternalValue iv = toInternalValue(constp.get(rFarg));
        if (iv.type != InternalValueType.UFUN_PROTOTYPE) {
            throw new InternalVMError("Expected UFunction-skeleton object.");
        }
        UserFunctionPrototype ufs = (UserFunctionPrototype) iv.getObject();
        UserFunction uf = ufs.createFunction(varp);
        varp.put(arg, new InternalValue(uf, InternalValueType.UFUNCTION));
    }

    private final void addName(int varId, int name) throws VMError {
        String nm;
        try {
            nm = constp.get(name).getString();
        } catch (InvalidValueTypeError ex) {
            throw new InternalVMError("Invalid function name in const pool.");
        }
        varp.addVar(varId, nm);
    }

    private final void removeName(int name) {
        varp.removeVar(name);
    }

    public List<Collection<Variable>> getVariables() throws VMError {
        return variables;
    }
}
