/*
 *  Fingy. Educational programming environment for children.
 *  Copyright (C) 2010.  Zhlobich Andrei <mailto:anjensan@informatics.by>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package by.informatics.jcx.fingy.lang.vm;

import java.util.AbstractCollection;
import java.util.Collection;
import java.util.Iterator;

import by.informatics.jcx.fingy.lang.Value;
import by.informatics.jcx.fingy.lang.Variable;

class VarPool {

    private class VPIterator implements Iterator<Variable> {

        private int index = 0;

        @Override
        public boolean hasNext() {
            shiftIndex();
            return index < length;
        }

        @Override
        public Variable next() {
            shiftIndex();
            return variables[index++];
        }

        @Override
        public void remove() {
            throw new UnsupportedOperationException();
        }
        
        private void shiftIndex() {
            while (index < length && variables[index] == null) {
                ++index;
            }
        }
    };

    private class VPCollection extends AbstractCollection<Variable> {
        @Override
        public Iterator<Variable> iterator() {
            return new VPIterator();
        }

        @Override
        public int size() {
            return varSize;
        }
    };

    public final int length;
    public final Value[] pool;
    public final VarPool parent;
    public final int shift;
    private VPVariable[] variables;
    private int varSize;
    private VPCollection varCol;

    public VarPool(int maxid) {
        this.length = maxid;
        this.pool = new Value[maxid];
        this.parent = null;
        this.shift = 0;
    }

    VarPool(int maxid, VarPool vp, int shift) {
        this.length = maxid;
        this.pool = new Value[maxid];
        this.parent = vp;
        this.shift = shift;
    }

    public final Value get(int id) throws InternalVMError {
        if (id >= shift) {
            return getAbsolute(id - shift);
        } else {
            return parent.get(id);
        }
    }

    public final void put(int id, Value value) throws InternalVMError {
        if (id >= shift) {
            putAbsolute(id - shift, value);
        } else {
            parent.put(id, value);
        }
    }

    public final Value getAbsolute(int id) throws InternalVMError {
        try {
            return pool[id];
        } catch (ArrayIndexOutOfBoundsException ex) {
            throw new InternalVMError("VarPool: invalid var id.");
        }
    }

    public final void putAbsolute(int id, Value value) throws InternalVMError {
        try {
            pool[id] = value;
            if (variables != null) {
                VPVariable v = getVar(id + shift);
                if (v != null) {
                    v.fireChange();
                }
            }
        } catch (ArrayIndexOutOfBoundsException ex) {
            throw new InternalVMError("VarPool: invalid var id.");
        }
    }

    public final void initVariables() {
        variables = new VPVariable[length];
        varCol = new VPCollection();
    }

    public final VPVariable addVar(int id, String name) {
        VPVariable v = new VPVariable(this, name, id);
        if (variables[id - shift] == null) {
            varSize++;
        }
        variables[id - shift] = v;
        return v;
    }

    public final VPVariable getVar(int id) {
        return variables[id - shift];
    }

    public final void removeVar(int id) {
        VPVariable v = variables[id - shift];
        if (v != null) {
            v.setValid(false);
            varSize--;
        }
        variables[id - shift] = null;
    }

    public Collection<? extends Variable> getVariables() {
        return varCol;
    }
}
